from __future__ import annotations

from gaphor import UML
from gaphor.core.modeling import Base, Diagram
from gaphor.diagram.drop import (
    diagram_has_presentation,
    drop,
    drop_relationship,
    grow_parent,
)
from gaphor.diagram.group import change_owner
from gaphor.diagram.presentation import ElementPresentation, Presentation, connect
from gaphor.diagram.support import get_diagram_item, get_diagram_item_metadata


@drop.register(UML.Element, Diagram)
def drop_element(
    element: UML.Element, diagram: Diagram, x: float, y: float
) -> Presentation | None:
    if item_class := get_diagram_item(type(element)):
        item = diagram.create(item_class)
        assert item

        item.matrix.translate(x, y)
        item.subject = element

        return item
    return None


@drop.register(UML.Association, Diagram)
def drop_association(element: UML.Association, diagram: Diagram, x, y):
    return drop_relationship(
        element, element.memberEnd[0].type, element.memberEnd[1].type, diagram, x, y
    )


@drop.register(UML.Connector, Diagram)
def drop_connector(element: UML.Connector, diagram: Diagram, x, y):
    return drop_relationship(
        element, element.end[0].role, element.end[1].role, diagram, x, y
    )


@drop.register(UML.Message, Diagram)
def drop_message(element: UML.Message, diagram: Diagram, x, y):
    return drop_relationship(
        element,
        element.sendEvent.covered
        if isinstance(element.sendEvent, UML.MessageOccurrenceSpecification)
        else None,
        element.receiveEvent.covered
        if isinstance(element.receiveEvent, UML.MessageOccurrenceSpecification)
        else None,
        diagram,
        x,
        y,
    )


@drop.register(UML.Pin, Diagram)
def drop_pin(element: UML.Pin, diagram: Diagram, x, y):
    return drop_pin_on_diagram(element, element.owner, diagram, x, y)


def drop_pin_on_diagram(
    element: Base, owner: Base | None, diagram: Diagram, x, y
) -> Presentation | None:
    item_class = get_diagram_item(type(element))
    if not item_class:
        return None

    owner_item = diagram_has_presentation(diagram, owner)
    if owner and not owner_item:
        return None

    item = diagram.create(item_class)
    assert item

    item.matrix.translate(x, y)
    item.subject = element
    handle = item.handles()[-1]
    if owner_item:
        connect(item, handle, owner_item)

    return item


@drop.register(UML.DirectedRelationship, Diagram)
def drop_relationship_on_diagram(
    element: UML.DirectedRelationship, diagram: Diagram, x: float, y: float
) -> Presentation | None:
    item_class = get_diagram_item(type(element))
    if not item_class:
        return None

    metadata = get_diagram_item_metadata(item_class)
    if metadata:
        # DirectedRelationships can be many-to-many, so we may need to create multiple items
        if metadata["head"].upper == 1 and metadata["tail"].upper == 1:
            head = metadata["head"].get(element)
            tail = metadata["tail"].get(element)
            new_item = drop_relationship(element, head, tail, diagram, x, y)
            if new_item:
                return new_item
        else:
            for head in metadata["head"].get(element):
                for tail in metadata["tail"].get(element):
                    new_item = drop_relationship(element, head, tail, diagram, x, y)
                    if new_item:
                        return new_item
    return None


@drop.register(UML.Element, ElementPresentation)
def drop_element_on_element_presentation(
    element: UML.Element, parent_item: ElementPresentation, x: float, y: float
) -> Presentation | None:
    item_class = get_diagram_item(type(element))
    if not item_class:
        return None

    diagram = parent_item.diagram
    item = diagram.create(item_class, subject=element)
    assert item

    if change_owner(parent_item.subject, element):
        item.parent = parent_item
        item.matrix.translate(x, y)
        grow_parent(parent_item, item)

    parent_item.request_update()

    return item



from gaphor.diagram.drop import drop, grow_parent
from gaphor.diagram.group import change_owner, ungroup
from gaphor.UML.states.pseudostates import PseudostateItem
from gaphor.UML.states.state import StateItem
from gaphor.UML.states.statemachine import StateMachineItem
from gaphor.UML.states.transition import TransitionItem


@drop.register(StateItem, StateItem)
@drop.register(StateItem, StateMachineItem)
@drop.register(PseudostateItem, StateItem)
@drop.register(PseudostateItem, StateMachineItem)
@drop.register(TransitionItem, StateItem)
@drop.register(TransitionItem, StateMachineItem)
def drop_region(item, new_parent: StateItem | StateMachineItem, x, y):
    assert item.diagram is new_parent.diagram

    if not item.subject:
        return

    old_parent = item.parent
    target_subject = new_parent.region_at_point((x, y))

    if target_subject is item.subject.container:
        return

    if old_parent and ungroup(old_parent.subject, item.subject):
        item.change_parent(None)
        old_parent.request_update()

    if change_owner(target_subject, item.subject):
        grow_parent(new_parent, item)
        item.change_parent(new_parent)



"""OR gate item definition."""

from math import pi

from gaphas.geometry import Rectangle

from gaphor.core.modeling import DrawContext
from gaphor.diagram.presentation import (
    Classified,
    ElementPresentation,
    text_name,
)
from gaphor.diagram.shapes import Box, IconBox, stroke
from gaphor.diagram.support import represents
from gaphor.RAAML import raaml
from gaphor.RAAML.fta.constants import DEFAULT_FTA_MAJOR, DEFAULT_FTA_MINOR
from gaphor.UML.compartments import text_from_package, text_stereotypes


@represents(raaml.OR)
class ORItem(Classified, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=DEFAULT_FTA_MINOR, height=DEFAULT_FTA_MAJOR)

        self.watch("subject[UML:NamedElement].name").watch(
            "subject[UML:NamedElement].namespace.name"
        )

    def update_shapes(self, event=None):
        self.shape = IconBox(
            Box(
                draw=draw_or_gate,
            ),
            text_stereotypes(self, lambda: [self.diagram.gettext("OR Gate")]),
            text_name(self),
            text_from_package(self),
        )


def draw_or_gate(box, context: DrawContext, bounding_box: Rectangle):
    cr = context.cairo
    left = 0
    right = bounding_box.width
    wall_top = bounding_box.height * 4.8 / 6.0
    wall_bottom = bounding_box.height * 4.5 / 5.0

    # Left wall
    cr.move_to(left, wall_bottom)
    cr.line_to(left, wall_top)

    # Right wall
    cr.move_to(right, wall_bottom)
    cr.line_to(right, wall_top)

    # Top left curve
    rx = right - left
    ry = bounding_box.height * 2.0 / 5.0
    cr.move_to(left, wall_top)
    point_top = bounding_box.height / 3.0 - ry / 2.0
    mid_width = bounding_box.width / 2.0
    cr.curve_to(
        left,
        bounding_box.height / 3.0,
        left + rx / 3.0,
        point_top,
        mid_width,
        point_top,
    )

    # Top right curve
    cr.move_to(right, wall_top)
    cr.curve_to(
        right,
        bounding_box.height / 3.0,
        right - rx / 3.0,
        point_top,
        mid_width,
        point_top,
    )

    # Bottom arc
    ry = bounding_box.height / 6.0
    cr.move_to(left, wall_bottom)
    cr.save()
    cr.translate(left + rx / 2.0, wall_bottom)
    cr.scale(rx / 2.0, ry / 2.0)
    cr.arc(0.0, 0.0, 1.0, pi, 0)
    cr.restore()

    # Bottom vertical lines
    left_line = left + rx / 5.0
    vertical_top = wall_bottom - ry / 2.4
    cr.move_to(left_line, vertical_top)
    cr.line_to(left_line, bounding_box.height)
    right_line = right - rx / 5.0
    cr.move_to(right_line, vertical_top)
    cr.line_to(right_line, bounding_box.height)

    # Top vertical line
    center = bounding_box.width / 2.0
    cr.move_to(center, point_top)
    cr.line_to(center, 0)
    stroke(context, fill=True)



"""Operational Situation item definition."""

from gaphas.geometry import Rectangle

from gaphor.core.modeling import DrawContext
from gaphor.diagram.presentation import (
    Classified,
    ElementPresentation,
)
from gaphor.diagram.shapes import Box, draw_border
from gaphor.diagram.support import represents
from gaphor.RAAML import raaml
from gaphor.UML.compartments import name_compartment


@represents(raaml.OperationalSituation)
@represents(raaml.AbstractOperationalSituation)
class OperationalSituationItem(Classified, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)

        self.watch("subject[UML:NamedElement].name").watch(
            "subject[UML:NamedElement].namespace.name"
        )

    def update_shapes(self, event=None):
        self.shape = Box(
            name_compartment(
                self, lambda: [self.diagram.gettext("Operational Situation")]
            ),
            draw=draw_operational_situation,
        )


def draw_operational_situation(box, context: DrawContext, bounding_box: Rectangle):
    draw_border(box, context, bounding_box)



import ast
import hashlib
import logging
from collections.abc import Iterator
from io import IOBase
from pathlib import Path

from gaphor import settings
from gaphor.abc import Service
from gaphor.core import event_handler
from gaphor.core.modeling import (
    AssociationAdded,
    AssociationDeleted,
    AssociationSet,
    AttributeUpdated,
    DerivedUpdated,
    ElementCreated,
    ElementDeleted,
    ElementTypeUpdated,
    ModelReady,
    RedefinedAdded,
    RedefinedDeleted,
    RedefinedSet,
    swap_element_type,
)
from gaphor.core.modeling.presentation import MatrixUpdated
from gaphor.diagram.connectors import (
    ItemConnected,
    ItemDisconnected,
    ItemReconnected,
    ItemTemporaryDisconnected,
)
from gaphor.diagram.presentation import HandlePositionEvent
from gaphor.diagram.segment import LineMergeSegmentEvent, LineSplitSegmentEvent
from gaphor.event import (
    ModelSaved,
    Notification,
    SessionCreated,
    SessionShutdown,
)
from gaphor.i18n import gettext
from gaphor.transaction import Transaction, TransactionCommit, TransactionRollback

log = logging.getLogger(__name__)


def sessions_dir() -> Path:
    d = settings.get_cache_dir() / "sessions"
    d.mkdir(exist_ok=True)
    return d


def all_sessions() -> Iterator[tuple[str, Path | None, Path | None]]:
    """Get all open active/inactive sessions.

    Returns a list of tuples: session id, filename path, template path.
    """
    for session_file in sessions_dir().glob("*.recovery"):
        with session_file.open(encoding="utf-8") as f:
            preamble_line = f.readline()
        if preamble_line:
            try:
                preamble = ast.literal_eval(preamble_line)
                path = Path(preamble.get("path"))
                is_template = preamble.get("template", False)
                if path.exists() and path.is_file():
                    yield (
                        (session_file.stem, None, Path(path))
                        if is_template
                        else (session_file.stem, Path(path), None)
                    )
                else:
                    log.info("Session file does not reference an existing model file.")
                    _move_aside(session_file)
            except (SyntaxError, TypeError, AttributeError):
                log.info("File has an invalid header: '%s'.", preamble_line)
                _move_aside(session_file)


class Recovery(Service):
    def __init__(self, event_manager, element_factory, modeling_language):
        self.event_manager = event_manager
        self.element_factory = element_factory
        self.modeling_language = modeling_language
        self.session_id: str = "_"
        self.recorder = Recorder()
        self.event_log: EventLog | None = None

        event_manager.subscribe(self.on_transaction_commit)
        event_manager.subscribe(self.on_transaction_rollback)
        event_manager.subscribe(self.on_model_loaded)
        event_manager.subscribe(self.on_model_ready)
        event_manager.subscribe(self.on_model_saved)
        event_manager.subscribe(self.on_session_shutdown)

        self.recorder.subscribe(event_manager)

    def shutdown(self):
        if self.event_log:
            self.event_log.close()

        self.event_manager.unsubscribe(self.on_transaction_commit)
        self.event_manager.unsubscribe(self.on_transaction_rollback)
        self.event_manager.unsubscribe(self.on_model_loaded)
        self.event_manager.unsubscribe(self.on_model_ready)
        self.event_manager.unsubscribe(self.on_model_saved)
        self.event_manager.unsubscribe(self.on_session_shutdown)

        self.recorder.unsubscribe(self.event_manager)

    @event_handler(TransactionCommit)
    def on_transaction_commit(self, event: TransactionCommit):
        if (
            self.event_log
            and self.recorder.events
            and event.context not in ("rollback", "recover")
        ):
            self.event_log.write(self.recorder.events)
        self.recorder.truncate()

    @event_handler(TransactionRollback)
    def on_transaction_rollback(self, _event):
        self.recorder.truncate()

    @event_handler(SessionCreated)
    def on_model_loaded(self, event: SessionCreated):
        self.session_id = event.session.session_id
        if event.filename or event.template:
            self.event_log = EventLog(self.session_id, event.filename, event.template)
        self.recorder.truncate()

    @event_handler(ModelReady)
    def on_model_ready(self, _event):
        if not self.event_log:
            return

        events = []
        try:
            with Transaction(self.event_manager, context="recover"):
                for events in self.event_log.read():
                    replay_events(events, self.element_factory, self.modeling_language)
        except Exception:
            log.error(
                "Could not recover model changes from %s. Changes have been rolled back.",
                self.event_log.log_file,
                exc_info=True,
            )
            log.warning("Replaying events failed.")
            self.event_log.move_aside()

        if events:
            self.event_manager.handle(
                Notification(
                    gettext(
                        "This model contains unsaved changes. The changes have been restored."
                    )
                )
            )

    @event_handler(ModelSaved)
    def on_model_saved(self, event: ModelSaved):
        if self.event_log:
            self.event_log.clear()

        if event.filename:
            self.event_log = EventLog(self.session_id, event.filename)
            self.event_log.clear()
        else:
            self.event_log = None

        self.recorder.truncate()

    @event_handler(SessionShutdown)
    def on_session_shutdown(self, _event: SessionShutdown):
        if self.event_log:
            self.event_log.clear()
        self.recorder.truncate()


class EventLog:
    def __init__(
        self, session_id: str, filename: Path | None, template: Path | None = None
    ):
        self._filename = filename
        self._template = template
        self._log_name = (sessions_dir() / session_id).with_suffix(".recovery")

        # The file that we use to save the events to:
        self._file: IOBase | None = None

    @property
    def log_file(self):
        return self._log_name

    def clear(self):
        self.close()
        self._log_name.unlink(missing_ok=True)

    def write(self, event):
        if not (self._filename or self._template):
            return

        f = self._file
        if not f or f.closed:
            f = self._file = self._log_name.open(mode="a", encoding="utf-8")

        if f.tell() == 0:
            if self._template:
                filename = self._template.absolute()
                is_template = True
            else:
                assert self._filename
                filename = self._filename.absolute()
                is_template = False

            f.write(
                repr(
                    {
                        "path": str(filename.absolute()),
                        "sha256": sha256sum(filename),
                        "template": is_template,
                    }
                )
            )
            f.write("\n")

        f.write(repr(event))
        f.write("\n")
        f.flush()

    def read(self):
        if not (self._filename or self._template):
            return

        self.close()
        try:
            with self._log_name.open(mode="r", encoding="utf-8") as f:
                preamble_line = f.readline()
                preamble = ast.literal_eval(preamble_line)
                if not isinstance(preamble, dict):
                    raise ChecksumFailed()

                filename = (
                    self._template if preamble.get("template") else self._filename
                )
                if not filename or sha256sum(filename) != preamble.get("sha256"):
                    raise ChecksumFailed()

                for line in f:
                    events = ast.literal_eval(line.rstrip("\r\n"))
                    yield events

        except FileNotFoundError:
            # Log does not exist, no problem
            pass
        except ChecksumFailed:
            # Move file after it's closed.
            log.info("Recovery file hash does not match.")
            self.move_aside()

    def close(self):
        if self._file:
            self._file.close()
            self._file = None

    def move_aside(self):
        self.close()
        _move_aside(self._log_name)


class ChecksumFailed(Exception):
    pass


def _move_aside(path: Path):
    backup = path.with_suffix(".recovery.bak")
    path.rename(backup)
    log.info("Session recovery file is renamed to %s.", backup)


def sha256sum(filename: Path):
    with open(filename, "rb", buffering=0) as f:
        return hashlib.file_digest(f, "sha256").hexdigest()


class Recorder:
    def __init__(self):
        self.events = []

    def subscribe(self, event_manager):
        event_manager.subscribe(self.on_create_element_event)
        event_manager.subscribe(self.on_delete_element_event)
        event_manager.subscribe(self.on_attribute_change_event)
        event_manager.subscribe(self.on_association_set_event)
        event_manager.subscribe(self.on_association_delete_event)
        event_manager.subscribe(self.on_matrix_updated)
        event_manager.subscribe(self.on_type_swapped_event)
        event_manager.subscribe(self.on_item_connected)
        event_manager.subscribe(self.on_item_disconnected)
        event_manager.subscribe(self.on_item_reconnected)
        event_manager.subscribe(self.on_handle_position_event)
        event_manager.subscribe(self.on_split_line_segment_event)
        event_manager.subscribe(self.on_merge_line_segment_event)

    def unsubscribe(self, event_manager):
        event_manager.unsubscribe(self.on_create_element_event)
        event_manager.unsubscribe(self.on_delete_element_event)
        event_manager.unsubscribe(self.on_attribute_change_event)
        event_manager.unsubscribe(self.on_association_set_event)
        event_manager.unsubscribe(self.on_association_delete_event)
        event_manager.unsubscribe(self.on_matrix_updated)
        event_manager.unsubscribe(self.on_type_swapped_event)
        event_manager.unsubscribe(self.on_item_connected)
        event_manager.unsubscribe(self.on_item_disconnected)
        event_manager.unsubscribe(self.on_item_reconnected)
        event_manager.unsubscribe(self.on_handle_position_event)
        event_manager.unsubscribe(self.on_split_line_segment_event)
        event_manager.unsubscribe(self.on_merge_line_segment_event)

    def truncate(self):
        del self.events[:]

    @event_handler(ElementCreated)
    def on_create_element_event(self, event: ElementCreated):
        self.events.append(
            (
                "c",
                type(event.element).__modeling_language__,
                type(event.element).__name__,
                event.element.id,
                event.diagram and event.diagram.id,
            )
        )

    @event_handler(ElementDeleted)
    def on_delete_element_event(self, event: ElementDeleted):
        self.events.append(("u", event.element.id, event.diagram and event.diagram.id))

    @event_handler(AttributeUpdated)
    def on_attribute_change_event(self, event: AttributeUpdated):
        self.events.append(
            ("a", event.element.id, event.property.name, event.new_value)
        )

    @event_handler(AssociationAdded, AssociationSet)
    def on_association_set_event(self, event: AssociationSet | AssociationAdded):
        if isinstance(event, DerivedUpdated | RedefinedAdded | RedefinedSet):
            return
        self.events.append(
            (
                "s",
                event.element.id,
                event.property.name,
                event.new_value and event.new_value.id,
            )
        )

    @event_handler(AssociationDeleted)
    def on_association_delete_event(self, event: AssociationDeleted):
        if isinstance(event, DerivedUpdated | RedefinedDeleted):
            return
        self.events.append(
            (
                "d",
                event.element.id,
                event.property.name,
                event.old_value and event.old_value.id,
            )
        )

    # We handle all RevertibleEvent subtypes separately:
    #
    # - gaphor.core.modeling.presentation.MatrixUpdated
    # - gaphor.diagram.presentation.HandlePositionEvent
    # - gaphor.diagram.connectors.ItemConnected
    # - gaphor.diagram.connectors.ItemDisconnected
    # - gaphor.diagram.connectors.ItemReconnected
    # - gaphor.diagram.tools.segment.LineSplitSegmentEvent
    # - gaphor.diagram.tools.segment.LineMergeSegmentEvent
    #
    # We can ignore gaphor.diagram.connectors.ItemTemporaryDisconnected,
    # since it has no (direct) impact on the model state.

    @event_handler(MatrixUpdated)
    def on_matrix_updated(self, event: MatrixUpdated):
        self.events.append(("mu", event.element.id, event.new_value))

    @event_handler(ElementTypeUpdated)
    def on_type_swapped_event(self, event: ElementTypeUpdated):
        self.events.append(("ts", event.element.id, event.new_class.__name__))

    @event_handler(HandlePositionEvent)
    def on_handle_position_event(self, event: HandlePositionEvent):
        self.events.append(
            ("hp", event.element.id, event.handle_index, event.new_value)
        )

    @event_handler(ItemConnected)
    def on_item_connected(self, event: ItemConnected):
        self.events.append(
            (
                "ic",
                event.element.id,
                event.handle_index,
                event.connected_id,
                event.port_index,
            )
        )

    @event_handler(ItemDisconnected)
    def on_item_disconnected(self, event: ItemDisconnected):
        self.events.append(
            (
                "id",
                event.element.id,
                event.handle_index,
                event.connected_id,
                event.port_index,
            )
        )

    @event_handler(ItemReconnected)
    def on_item_reconnected(self, event: ItemReconnected):
        self.events.append(
            (
                "ir",
                event.element.id,
                event.handle_index,
                event.connected_id,
                event.port_index,
            )
        )

    @event_handler(LineSplitSegmentEvent)
    def on_split_line_segment_event(self, event: LineSplitSegmentEvent):
        self.events.append(
            (
                "ls",
                event.element.id,
                event.segment,
                event.count,
            )
        )

    @event_handler(LineMergeSegmentEvent)
    def on_merge_line_segment_event(self, event: LineMergeSegmentEvent):
        self.events.append(
            (
                "lm",
                event.element.id,
                event.segment,
                event.count,
            )
        )


def replay_events(events, element_factory, modeling_language):
    """Replay events previously recorded by EventLog."""
    for event in events:
        match event:
            case ("c", ns, type, element_id, None):
                element_factory.create_as(
                    modeling_language.lookup_element(type, ns), element_id
                )
            case ("c", ns, type, element_id, diagram_id):
                diagram = element_factory.lookup(diagram_id)
                diagram.create_as(
                    modeling_language.lookup_element(type, ns), element_id
                )
            case ("u", element_id, _diagram_id):
                element_factory.lookup(element_id).unlink()
            case ("a", element_id, prop, value):
                element = element_factory.lookup(element_id)
                setattr(element, prop, value)
            case ("s", element_id, prop, other_element_id):
                element = element_factory.lookup(element_id)
                other_element = element_factory.lookup(other_element_id)
                setattr(element, prop, other_element)
            case ("d", element_id, prop, other_element_id):
                element = element_factory.lookup(element_id)
                other_element = element_factory.lookup(other_element_id)
                del getattr(element, prop)[other_element]
            case ("mu", element_id, matrix):
                element = element_factory.lookup(element_id)
                element.matrix.set(*matrix)
            case ("hp", element_id, handle_index, pos):
                element = element_factory.lookup(element_id)
                element.handles()[handle_index].pos = pos
            case ("ic", element_id, handle_index, connected_id, port_index):
                element = element_factory.lookup(element_id)
                connected = element_factory.lookup(connected_id)
                ItemDisconnected(
                    element,
                    element.handles()[handle_index],
                    connected,
                    connected.ports()[port_index],
                ).revert(element)
            case ("id", element_id, handle_index, connected_id, port_index):
                element = element_factory.lookup(element_id)
                connected = element_factory.lookup(connected_id)
                ItemConnected(
                    element,
                    element.handles()[handle_index],
                    connected,
                    connected.ports()[port_index],
                ).revert(element)
            case ("ir", element_id, handle_index, connected_id, port_index):
                element = element_factory.lookup(element_id)
                connected = element_factory.lookup(connected_id)
                ItemTemporaryDisconnected(
                    element,
                    element.handles()[handle_index],
                    connected,
                    connected.ports()[port_index],
                ).revert(element)
            case ("ls", element_id, segment, count):
                element = element_factory.lookup(element_id)
                LineMergeSegmentEvent(element, segment, count).revert(element)
            case ("lm", element_id, segment, count):
                element = element_factory.lookup(element_id)
                LineSplitSegmentEvent(element, segment, count).revert(element)
            case ("ts", element_id, type):
                element = element_factory.lookup(element_id)
                swap_element_type(element, modeling_language.lookup_element(type))
            case _:
                assert NotImplementedError(f"Event {event} not implemented")



"""The action definition for the UML toolbox."""

import gaphor.UML.uml as UML
from gaphor.diagram.diagramtoolbox import (
    DiagramType,
    DiagramTypes,
    ElementCreateInfo,
    ToolboxDefinition,
)
from gaphor.i18n import i18nize
from gaphor.UML.actions.actionstoolbox import actions
from gaphor.UML.classes.classestoolbox import classes
from gaphor.UML.deployments.deploymentstoolbox import deployments
from gaphor.UML.general.generaltoolbox import general_tools
from gaphor.UML.interactions.interactionstoolbox import interactions
from gaphor.UML.profiles.profilestoolbox import profiles
from gaphor.UML.states.statestoolbox import states
from gaphor.UML.usecases.usecasetoolbox import use_cases

uml_toolbox_actions: ToolboxDefinition = (
    general_tools,
    classes,
    deployments,
    actions,
    interactions,
    states,
    use_cases,
    profiles,
)

uml_diagram_types: DiagramTypes = (
    DiagramType(UML.ClassDiagram, i18nize("Class Diagram"), (classes,)),
    DiagramType(UML.PackageDiagram, i18nize("Package Diagram"), (classes,)),
    DiagramType(UML.ComponentDiagram, i18nize("Component Diagram"), (classes,)),
    DiagramType(UML.DeploymentDiagram, i18nize("Deployment Diagram"), (deployments,)),
    DiagramType(UML.ActivityDiagram, i18nize("Activity Diagram"), (actions,)),
    DiagramType(UML.SequenceDiagram, i18nize("Sequence Diagram"), (interactions,)),
    DiagramType(
        UML.CommunicationDiagram, i18nize("Communication Diagram"), (interactions,)
    ),
    DiagramType(UML.StateMachineDiagram, i18nize("State Machine Diagram"), (states,)),
    DiagramType(UML.UseCaseDiagram, i18nize("Use Case Diagram"), (use_cases,)),
    DiagramType(UML.ProfileDiagram, i18nize("Profile Diagram"), (profiles,)),
    DiagramType(UML.Diagram, i18nize("Generic Diagram"), ()),
)

uml_element_types = (
    ElementCreateInfo("package", i18nize("Package"), UML.Package, (UML.Package,)),
    ElementCreateInfo("activity", i18nize("Activity"), UML.Activity, (UML.Package,)),
    ElementCreateInfo("actor", i18nize("Actor"), UML.Actor, (UML.Package,)),
    ElementCreateInfo("artifact", i18nize("Artifact"), UML.Artifact, (UML.Package,)),
    ElementCreateInfo("class", i18nize("Class"), UML.Class, (UML.Package,)),
    ElementCreateInfo("component", i18nize("Component"), UML.Component, (UML.Package,)),
    ElementCreateInfo("datatype", i18nize("Data Type"), UML.DataType, (UML.Package,)),
    ElementCreateInfo("device", i18nize("Device"), UML.Device, (UML.Package,)),
    ElementCreateInfo(
        "enumeration", i18nize("Enumeration"), UML.Enumeration, (UML.Package,)
    ),
    ElementCreateInfo(
        "instancespecification",
        i18nize("Instance Specification"),
        UML.InstanceSpecification,
        (UML.Package,),
    ),
    ElementCreateInfo(
        "interaction", i18nize("Interaction"), UML.Interaction, (UML.Package,)
    ),
    ElementCreateInfo("interface", i18nize("Interface"), UML.Interface, (UML.Package,)),
    ElementCreateInfo("node", i18nize("Node"), UML.Node, (UML.Package,)),
    ElementCreateInfo("profile", i18nize("Profile"), UML.Profile, (UML.Package,)),
    ElementCreateInfo(
        "statemachine", i18nize("State Machine"), UML.StateMachine, (UML.Package,)
    ),
    ElementCreateInfo(
        "stereotype", i18nize("Stereotype"), UML.Stereotype, (UML.Package,)
    ),
    ElementCreateInfo("usecase", i18nize("Use Case"), UML.UseCase, (UML.Package,)),
)



"""Event Manager."""

import asyncio
import inspect
from collections import deque

from generic.event import Event, Handler
from generic.event import Manager as _Manager

from gaphor.abc import Service


def event_handler(*event_types):
    """Mark a function/method as an event handler for a particular type of
    event.

    Given a custom event type:

    >>> class CustomEvent:
    ...     def __str__(self):
    ...         return type(self).__name__

    You can apply this to a handler method or function:

    >>> @event_handler(CustomEvent)
    ... def custom_handler(event: CustomEvent):
    ...     print(event)

    This will allow you to let the even be handled by an event manager:

    >>> event_manager = EventManager()
    >>> event_manager.subscribe(custom_handler)
    >>> event_manager.handle(CustomEvent())
    CustomEvent
    """

    def wrapper(func):
        func.__event_types__ = event_types
        return func

    return wrapper


class EventManager(Service):
    """The Event Manager provides a flexible way to dispatch events.

    Event dispatching is a central component in Gaphor. It allows components
    in Gaphor to react to changes in the application.

    Events are dispatched by type.
    """

    def __init__(self) -> None:
        self._events = _Manager()
        self._priority = _Manager()
        self._queue: deque[Event] = deque()
        self._async_tasks: set[asyncio.Task] = set()
        self._handling = False

    def shutdown(self) -> None:
        for task in set(self._async_tasks):
            task.cancel()

    def subscribe(self, handler: Handler) -> None:
        """Register a handler.

        Handlers are triggered (executed) when events are
        emitted through the :obj:`~gaphor.core.eventmanager.EventManager.handle` method.
        """
        if inspect.iscoroutinefunction(handler):

            def async_handler(event: Event) -> None:
                coro = handler(event)
                task = asyncio.create_task(coro)
                self._async_tasks.add(task)
                task.add_done_callback(self._task_done)

            self._subscribe(
                self._events, async_handler, getattr(handler, "__event_types__", None)
            )
        else:
            self._subscribe(self._events, handler)

    async def gather_tasks(self):
        if self._async_tasks:
            await asyncio.gather(*self._async_tasks)

    def _task_done(self, task: asyncio.Task) -> None:
        self._async_tasks.remove(task)

    def priority_subscribe(self, handler: Handler) -> None:
        """Register a handler.

        Priority handlers are executed directly. They should not raise
        other events, cause that can cause a problem in the exection order.

        It's basically to make sure that all events are recorded by the
        undo manager.
        """
        self._subscribe(self._priority, handler)

    def _subscribe(self, manager: _Manager, handler: Handler, event_types=None) -> None:
        event_types = getattr(handler, "__event_types__", event_types)
        if not event_types:
            raise Exception(f"No event types provided for function {handler}")

        for et in event_types:
            manager.subscribe(handler, et)

    def unsubscribe(self, handler: Handler) -> None:
        """Unregister a previously registered handler."""
        event_types = getattr(handler, "__event_types__", None)
        if not event_types:
            raise Exception(f"No event types provided for function {handler}")

        for et in event_types:
            self._priority.unsubscribe(handler, et)
            self._events.unsubscribe(handler, et)

    def handle(self, *events: Event) -> None:
        """Send event notifications to registered handlers."""
        queue = self._queue
        queue.extendleft(events)

        for event in events:
            self._priority.handle(event)

        if not self._handling:
            self._handling = True
            try:
                while queue:
                    self._events.handle(queue.pop())
            finally:
                self._handling = False



"""Pseudostate diagram items.

See also gaphor.UML.states package description.
"""

from gaphas.item import SE

from gaphor import UML
from gaphor.core.modeling.properties import relation_one
from gaphor.diagram.presentation import ElementPresentation, Named, text_name
from gaphor.diagram.shapes import Box, IconBox, ellipse, stroke
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_stereotypes


@represents(UML.Pseudostate)
class PseudostateItem(ElementPresentation, Named):
    """A Pseudostate:

    * initial (solid circle, in: 0, out: 1)
    * deepHistory (circle with "H*", in: *, out: 1)
    * shallowHistory (circle with "H", in: *, out: 1)
    * join (bar, in: *, out: 1)
    * fork (bar, in: 1, out: *)
    * junction (solid circle, in: *, out: *)
    * choice (diamond, in: *, out: *)
    * entryPoint (circle, in: *, out: *)
    * exitPoint (circle with cross, in: *, out: *)
    * terminate (cross, in: *, out: *)
    """

    subject: relation_one[UML.Pseudostate]

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=6, height=20)
        for h in self.handles():
            h.movable = False

        self.watch("subject[NamedElement].name")
        self.watch("subject[Element].appliedStereotype.classifier.name")
        self.watch("subject[Pseudostate].kind", self.update_shapes)

    def update_shapes(self, event=None):
        kind = self.subject.kind if self.subject and self.subject.kind else "initial"
        draw, width, height = PSEUDOSTATE_SHAPE[kind]
        self.handles()[SE].pos = (width, height)

        self.shape = IconBox(
            Box(draw=draw),
            text_stereotypes(self),
            text_name(self),
        )


def draw_initial_or_junction_pseudostate(box, context, bounding_box):
    """Draw initial pseudostate symbol."""
    cr = context.cairo
    if stroke := context.style["color"]:
        cr.set_source_rgba(*stroke)
    ellipse(cr, *bounding_box)
    cr.set_line_width(0.01)
    cr.fill()


def draw_deep_history_pseudostate(box, context, bounding_box):
    cr = context.cairo
    ellipse(cr, *bounding_box)
    cr.move_to(9, 10)
    cr.line_to(9, 20)
    cr.move_to(15, 10)
    cr.line_to(15, 20)
    cr.move_to(9, 15)
    cr.line_to(15, 15)

    for p in ((21, 9), (25, 12), (18.5, 17), (23.5, 17), (17, 12)):
        cr.move_to(21, 13)
        cr.line_to(*p)
    stroke(context, fill=False)


def draw_shallow_history_pseudostate(box, context, bounding_box):
    cr = context.cairo
    ellipse(cr, *bounding_box)
    cr.move_to(12, 10)
    cr.line_to(12, 20)
    cr.move_to(18, 10)
    cr.line_to(18, 20)
    cr.move_to(12, 15)
    cr.line_to(18, 15)
    stroke(context, fill=False)


def draw_join_or_fork_pseudostate(box, context, bounding_box):
    cr = context.cairo
    if stroke := context.style.get("color"):
        cr.set_source_rgba(*stroke)

    cr.set_line_width(6)
    cr.move_to(3, 0)
    cr.line_to(3, 40)
    cr.stroke()


def draw_choice_pseudostate(box, context, bounding_box):
    cr = context.cairo
    cr.move_to(15, 0)
    cr.line_to(30, 15)
    cr.line_to(15, 30)
    cr.line_to(0, 15)
    cr.close_path()
    stroke(context, fill=False)


def draw_entry_point_pseudostate(box, context, bounding_box):
    ellipse(context.cairo, *bounding_box)


def draw_exit_point_pseudostate(box, context, bounding_box):
    cr = context.cairo
    ellipse(cr, *bounding_box)
    cr.move_to(5, 5)
    cr.line_to(25, 25)
    cr.move_to(5, 25)
    cr.line_to(25, 5)
    stroke(context, fill=False)


def draw_terminate_pseudostate(box, context, bounding_box):
    cr = context.cairo
    cr.move_to(0, 0)
    cr.line_to(30, 20)
    cr.move_to(0, 20)
    cr.line_to(30, 0)
    stroke(context, fill=False)


PSEUDOSTATE_SHAPE = {
    "initial": (draw_initial_or_junction_pseudostate, 20, 20),
    "deepHistory": (draw_deep_history_pseudostate, 30, 30),
    "shallowHistory": (draw_shallow_history_pseudostate, 30, 30),
    "join": (draw_join_or_fork_pseudostate, 6, 40),
    "fork": (draw_join_or_fork_pseudostate, 6, 40),
    "junction": (draw_initial_or_junction_pseudostate, 20, 20),
    "choice": (draw_choice_pseudostate, 30, 30),
    "entryPoint": (draw_entry_point_pseudostate, 30, 30),
    "exitPoint": (draw_exit_point_pseudostate, 30, 30),
    "terminate": (draw_terminate_pseudostate, 30, 20),
}



"""The definition for the states section of the toolbox."""

from functools import partial

from gaphas.item import SE

from gaphor import UML
from gaphor.core import gettext
from gaphor.diagram.diagramtoolbox import ToolDef, ToolSection, new_item_factory
from gaphor.UML import diagramitems
from gaphor.UML.recipes import owner_package
from gaphor.UML.toolboxconfig import namespace_config


def state_config(new_item):
    state_machine_config(new_item, name=new_item.diagram.gettext("State"))


def pseudostate_config(new_item, kind):
    new_item.subject.kind = kind
    state_machine_config(new_item)


def state_machine_config(new_item, name=None):
    subject = new_item.subject
    if name:
        subject.name = new_item.diagram.gettext("New {name}").format(name=name)
    if subject.container:
        return

    diagram = new_item.diagram
    package = owner_package(diagram.owner)

    state_machines = (
        [i for i in package.ownedType if isinstance(i, UML.StateMachine)]
        if package
        else diagram.model.lselect(
            lambda e: isinstance(e, UML.StateMachine) and e.package is None
        )
    )

    if state_machines:
        state_machine = state_machines[0]
    else:
        state_machine = subject.model.create(UML.StateMachine)
        state_machine.name = new_item.diagram.gettext("State Machine")
        state_machine.package = package

    if state_machine.region:
        region = state_machine.region[0]
    else:
        region = subject.model.create(UML.Region)
        region.stateMachine = state_machine

    subject.container = region


states = ToolSection(
    gettext("States"),
    (
        ToolDef(
            "toolbox-state-machine",
            gettext("State Machine"),
            "gaphor-state-machine-symbolic",
            None,
            new_item_factory(
                diagramitems.StateMachineItem,
                UML.StateMachine,
                config_func=namespace_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-state",
            gettext("State"),
            "gaphor-state-symbolic",
            "s",
            new_item_factory(
                diagramitems.StateItem,
                UML.State,
                config_func=state_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-initial-pseudostate",
            gettext("Initial Pseudostate"),
            "gaphor-initial-pseudostate-symbolic",
            None,
            new_item_factory(
                diagramitems.PseudostateItem,
                UML.Pseudostate,
                partial(pseudostate_config, kind="initial"),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-final-state",
            gettext("Final State"),
            "gaphor-final-state-symbolic",
            None,
            new_item_factory(
                diagramitems.FinalStateItem,
                UML.FinalState,
                config_func=state_machine_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-transition",
            gettext("Transition"),
            "gaphor-transition-symbolic",
            "<Shift>T",
            new_item_factory(diagramitems.TransitionItem),
        ),
        ToolDef(
            "toolbox-shallow-history-pseudostate",
            gettext("Shallow History Pseudostate"),
            "gaphor-shallow-history-pseudostate-symbolic",
            None,
            new_item_factory(
                diagramitems.PseudostateItem,
                UML.Pseudostate,
                partial(pseudostate_config, kind="shallowHistory"),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-deep-history-pseudostate",
            gettext("Deep History Pseudostate"),
            "gaphor-deep-history-pseudostate-symbolic",
            None,
            new_item_factory(
                diagramitems.PseudostateItem,
                UML.Pseudostate,
                partial(pseudostate_config, kind="deepHistory"),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-join-pseudostate",
            gettext("Join Pseudostate"),
            "gaphor-join-pseudostate-symbolic",
            None,
            new_item_factory(
                diagramitems.PseudostateItem,
                UML.Pseudostate,
                partial(pseudostate_config, kind="join"),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-fork-pseudostate",
            gettext("Fork Pseudostate"),
            "gaphor-fork-pseudostate-symbolic",
            None,
            new_item_factory(
                diagramitems.PseudostateItem,
                UML.Pseudostate,
                partial(pseudostate_config, kind="fork"),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-junction-pseudostate",
            gettext("Junction Pseudostate"),
            "gaphor-junction-pseudostate-symbolic",
            None,
            new_item_factory(
                diagramitems.PseudostateItem,
                UML.Pseudostate,
                partial(pseudostate_config, kind="junction"),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-choice-pseudostate",
            gettext("Choice Pseudostate"),
            "gaphor-choice-pseudostate-symbolic",
            None,
            new_item_factory(
                diagramitems.PseudostateItem,
                UML.Pseudostate,
                partial(pseudostate_config, kind="choice"),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-entry-point-pseudostate",
            gettext("Entry Point Pseudostate"),
            "gaphor-entry-point-pseudostate-symbolic",
            None,
            new_item_factory(
                diagramitems.PseudostateItem,
                UML.Pseudostate,
                partial(pseudostate_config, kind="entryPoint"),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-exit-point-pseudostate",
            gettext("Exit Point Pseudostate"),
            "gaphor-exit-point-pseudostate-symbolic",
            None,
            new_item_factory(
                diagramitems.PseudostateItem,
                UML.Pseudostate,
                partial(pseudostate_config, kind="exitPoint"),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-terminate-pseudostate",
            gettext("Terminate Pseudostate"),
            "gaphor-terminate-pseudostate-symbolic",
            None,
            new_item_factory(
                diagramitems.PseudostateItem,
                UML.Pseudostate,
                partial(pseudostate_config, kind="terminate"),
            ),
            handle_index=SE,
        ),
    ),
)



from __future__ import annotations

from gaphas.types import Pos

from gaphor import UML
from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import Classified, ElementPresentation, text_name
from gaphor.diagram.shapes import Box, CssNode, draw_border
from gaphor.diagram.support import represents
from gaphor.UML.classes.stereotype import stereotype_compartments, stereotype_watches
from gaphor.UML.compartments import text_stereotypes
from gaphor.UML.states.region import region_compartment


@represents(UML.StateMachine)
class StateMachineItem(Classified, ElementPresentation[UML.StateMachine]):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)
        self._region_heights = []
        self.watch("children", self.update_shapes)
        self.watch("show_stereotypes", self.update_shapes)
        self.watch("show_regions", self.update_shapes)
        self.watch("subject[NamedElement].name")
        self.watch("subject[StateMachine].region.name")
        self.watch("subject[StateMachine].region", self.update_shapes)
        stereotype_watches(self)

    show_stereotypes: attribute[int] = attribute("show_stereotypes", int)
    show_regions: attribute[int] = attribute("show_regions", int, default=True)

    def update_shapes(self, event=None):
        region_boxes = (
            list(region_compartment(self.subject)) if self.show_regions else []
        )
        self.shape = Box(
            CssNode(
                "compartment",
                None,
                Box(
                    text_stereotypes(
                        self, lambda: [self.diagram.gettext("statemachine")]
                    ),
                    text_name(self),
                ),
            ),
            *(self.show_stereotypes and stereotype_compartments(self.subject) or []),
            *(
                region_boxes
                and [
                    CssNode(
                        "regions",
                        None,
                        Box(
                            *(region_boxes),
                        ),
                    )
                ]
                or []
            ),
            draw=draw_border,
        )

    def update(self, context):
        super().update(context)
        self._region_heights = [h for _w, h in self._shape.children[-1].child.sizes]  # type: ignore[union-attr]

    def region_at_point(self, pos: Pos) -> UML.Region | UML.StateMachine | None:
        region_offset = self.height - sum(self._region_heights)
        _x, y = pos
        for region, h in zip(self.subject.region, self._region_heights, strict=False):
            region_offset += h
            if y < region_offset:
                return region
        return self.subject



from gi.repository import Gio

from gaphor import UML
from gaphor.diagram.propertypages import (
    LabelValue,
    PropertyPageBase,
    PropertyPages,
    new_resource_builder,
)
from gaphor.transaction import Transaction
from gaphor.UML.actions.partition import PartitionItem

new_builder = new_resource_builder("gaphor.UML.actions")


@PropertyPages.register(PartitionItem)
class PartitionPropertyPage(PropertyPageBase):
    """Partition property page."""

    order = 15

    def __init__(self, item: PartitionItem, event_manager):
        super().__init__()
        self.item = item
        self.event_manager = event_manager
        self.partitions = None

    def construct(self):
        """Creates the Partition Property Page."""
        builder = new_builder(
            "partition-editor",
            "num-partitions-adjustment",
            signals={"partitions-changed": (self._on_num_partitions_changed,)},
        )

        num_partitions = builder.get_object("num-partitions")
        num_partitions.set_value(len(self.item.partition))

        self.partitions = builder.get_object("partitions")
        if self.partitions:
            for partition in self.item.partition:
                self.partitions.append(self.construct_partition(partition))

        return builder.get_object("partition-editor")

    def construct_partition(self, partition):
        builder = new_builder(
            "partition",
            signals={
                "partition-name-changed": (self._on_partition_name_changed, partition),
            },
        )

        builder.get_object("partition-name").set_text(partition.name or "")

        dropdown = builder.get_object("partition-type")
        model = Gio.ListStore.new(LabelValue)
        model.append(LabelValue("", None))
        for c in sorted(
            (
                c
                for c in self.item.model.select(UML.Classifier)
                if c.name and not isinstance(c, UML.Behavior)
            ),
            key=lambda c: c.name or "",
        ):
            model.append(LabelValue(c.name, c.id))
        dropdown.set_model(model)

        if partition.represents:
            dropdown.set_selected(
                next(
                    n
                    for n, lv in enumerate(model)
                    if lv.value == partition.represents.id
                )
            )

        dropdown.connect("notify::selected", self._on_partition_type_changed, partition)

        return builder.get_object("partition")

    def _on_num_partitions_changed(self, spin_button):
        """Event handler for partition number spin button."""
        num_partitions = spin_button.get_value_as_int()
        with Transaction(self.event_manager):
            self.update_partitions(num_partitions)

    def _on_partition_name_changed(self, entry, partition):
        """Event handler for editing partition names."""
        with Transaction(self.event_manager, context="editing"):
            partition.name = entry.get_text()

    def _on_partition_type_changed(self, combo, _pspec, partition):
        """Event handler for editing partition names."""
        with Transaction(self.event_manager):
            if id := combo.get_selected_item().value:
                element = self.item.model.lookup(id)
                partition.represents = element
            else:
                del partition.represents

    def update_partitions(self, num_partitions) -> None:
        """Add and remove partitions.

        Clear the list store, then add or remove UML.ActivityPartitions
        to account for updates in the number of partitions, and finally
        rebuild the list store.
        """
        if not self.item.subject or not self.partitions:
            return

        while num_partitions > len(self.item.partition):
            partition = self.item.subject.model.create(UML.ActivityPartition)
            partition.name = self.item.diagram.gettext("New Swimlane")
            partition.activity = self.item.subject.activity
            self.item.partition.append(partition)

            self.partitions.append(self.construct_partition(partition))
        while num_partitions < len(self.item.partition):
            partition = self.item.partition[-1]
            partition.unlink()

            last_child = self.partitions.get_last_child()
            self.partitions.remove(last_child)

        diagram = self.item.diagram
        diagram.update(diagram.ownedPresentation)



"""Message item connection adapters."""

from gaphor import UML
from gaphor.core.modeling import Presentation
from gaphor.diagram.connectors import BaseConnector, Connector
from gaphor.diagram.group import change_owner
from gaphor.UML.interactions.executionspecification import ExecutionSpecificationItem
from gaphor.UML.interactions.interaction import InteractionItem
from gaphor.UML.interactions.lifeline import LifelineItem
from gaphor.UML.interactions.message import MessageItem


def get_connected(item, handle) -> Presentation[UML.Element] | None:
    """Get item connected to a handle."""
    if cinfo := item.diagram.connections.get_connection(handle):
        return cinfo.connected  # type: ignore[no-any-return] # noqa: F723
    return None


def get_lifeline(item, handle):
    connected_item = get_connected(item, handle)
    if connected_item is None or isinstance(connected_item, LifelineItem):
        return connected_item
    return get_lifeline(connected_item, connected_item.handles()[0])


def order_lifeline_covered_by(lifeline):
    diagram = lifeline.diagram

    def y_and_occurrence(connected):
        for conn in diagram.connections.get_connections(connected=connected):
            m = conn.item.matrix_i2c
            if not conn.item.subject:
                # Can happen during DnD
                continue
            if isinstance(conn.item, ExecutionSpecificationItem):
                yield (
                    m.transform_point(*conn.item.top.pos)[1],
                    conn.item.subject.start,
                )
                yield (
                    m.transform_point(*conn.item.bottom.pos)[1],
                    conn.item.subject.finish,
                )
                yield from y_and_occurrence(conn.item)
            elif isinstance(conn.item, MessageItem):
                yield (
                    m.transform_point(*conn.handle.pos)[1],
                    conn.item.subject.sendEvent
                    if conn.handle is conn.item.head
                    else conn.item.subject.receiveEvent,
                )

    if lifeline.subject:
        keys = {o: y for y, o in y_and_occurrence(lifeline)}
        lifeline.subject.coveredBy.order(lambda key: keys.get(key, 0.0))


def owner_for_message(line, lifeline):
    maybe_interaction = lifeline.parent
    if line.subject.interaction:
        return
    elif isinstance(maybe_interaction, InteractionItem):
        line.parent = maybe_interaction
        change_owner(maybe_interaction, line)
    elif lifeline.subject and lifeline.subject.interaction:
        line.subject.interaction = lifeline.subject.interaction


def connect_lifelines(line, send, received):
    """Always create a new Message with two EventOccurrence instances."""

    def get_subject():
        if not line.subject:
            message = line.model.create(UML.Message)
            message.name = line.diagram.gettext("call()")
            line.subject = message
        return line.subject

    if send:
        message = get_subject()
        if not message.sendEvent:
            event = message.model.create(UML.MessageOccurrenceSpecification)
            event.sendMessage = message
            event.covered = send.subject
            order_lifeline_covered_by(send)
        owner_for_message(line, send)

    if received:
        message = get_subject()
        if not message.receiveEvent:
            event = message.model.create(UML.MessageOccurrenceSpecification)
            event.receiveMessage = message
            event.covered = received.subject
            order_lifeline_covered_by(received)
        owner_for_message(line, received)


def disconnect_lifelines(line, send, received):
    """Disconnect lifeline and set appropriate kind of message item.

    If there are no lifelines connected on both ends, then remove the
    message from the data model.
    """
    if not (subject := line.subject):
        return

    if send and (event := subject.sendEvent):
        event.unlink()

    if received and (event := subject.receiveEvent):
        event.unlink()

    # one is disconnected and one is about to be disconnected,
    # so destroy the message
    if not (subject.sendEvent or subject.receiveEvent):
        # Both ends are disconnected: remove subject
        del line.subject


@Connector.register(LifelineItem, MessageItem)
class MessageLifelineConnect(BaseConnector):
    """Connect lifeline with a message.

    A message can connect to both the lifeline's head (the rectangle) or
    the lifetime line. In case it's added to the head, the message is
    considered to be part of a communication diagram. If the message is
    added to a lifetime line, it's considered a sequence diagram.

    Message.head has no arrow (sendEvent)
    Message.tail has the arrow (receiveEvent)
    """

    element: LifelineItem
    line: MessageItem

    def connect(self, handle, port):
        line = self.line
        send = get_lifeline(line, line.head)
        received = get_lifeline(line, line.tail)
        connect_lifelines(line, send, received)
        return True

    def disconnect(self, handle):
        line = self.line
        send: Presentation[UML.Element] | None = get_connected(line, line.head)
        received = self.get_connected(line.tail)

        # if a message is a deleted message, then the disconnection causes
        # the lifeline to be no longer destroyed (note that there can be
        # only one delete message connected to lifeline)
        if received and line.subject and line.subject.messageSort == "deleteMessage":
            assert isinstance(received, LifelineItem)
            received.is_destroyed = False
            received.request_update()

        disconnect_lifelines(
            line,
            send and handle is line.head,
            received and handle is line.tail,
        )


@Connector.register(ExecutionSpecificationItem, MessageItem)
class ExecutionSpecificationMessageConnect(BaseConnector):
    element: ExecutionSpecificationItem
    line: MessageItem

    def connect(self, handle, _port):
        line = self.line
        send = get_lifeline(line, line.head)
        received = get_lifeline(line, line.tail)
        connect_lifelines(line, send, received)
        return True

    def disconnect(self, handle):
        line = self.line
        send = get_lifeline(line, line.head)
        received = get_lifeline(line, line.tail)
        disconnect_lifelines(
            line,
            send and handle is line.head,
            received and handle is line.tail,
        )


@Connector.register(LifelineItem, ExecutionSpecificationItem)
class LifelineExecutionSpecificationConnect(BaseConnector):
    element: LifelineItem
    line: ExecutionSpecificationItem

    def allow(self, handle, port):
        lifetime = self.element.lifetime
        return super().allow(handle, port) and lifetime.visible

    def connect(self, handle, port):
        lifeline = self.element.subject
        exec_spec = self.line.subject
        model = self.element.model
        if not exec_spec:
            exec_spec = model.create(UML.BehaviorExecutionSpecification)
            self.line.subject = exec_spec

            start_occurrence: UML.ExecutionOccurrenceSpecification = model.create(
                UML.ExecutionOccurrenceSpecification
            )
            start_occurrence.covered = lifeline
            start_occurrence.execution = exec_spec

            finish_occurrence: UML.ExecutionOccurrenceSpecification = model.create(
                UML.ExecutionOccurrenceSpecification
            )
            finish_occurrence.covered = lifeline
            finish_occurrence.execution = exec_spec

        if lifeline.interaction:
            exec_spec.enclosingInteraction = lifeline.interaction

        if self.line.parent is not self.element:
            self.line.change_parent(self.element)

        for cinfo in self.diagram.connections.get_connections(connected=self.line):
            Connector(self.line, cinfo.item).connect(cinfo.handle, cinfo.port)
        return True

    def disconnect(self, handle):
        del self.line.subject

        if self.line.parent is self.element:
            new_parent = self.element.parent
            self.line.change_parent(new_parent)

        for cinfo in self.diagram.connections.get_connections(connected=self.line):
            Connector(self.line, cinfo.item).disconnect(cinfo.handle)


@Connector.register(ExecutionSpecificationItem, ExecutionSpecificationItem)
class ExecutionSpecificationExecutionSpecificationConnect(BaseConnector):
    element: ExecutionSpecificationItem
    line: ExecutionSpecificationItem

    def connect(self, handle, _port):
        parent_exec_spec = self.element.subject

        if not parent_exec_spec:
            # Can connect child exec spec if parent is not connected
            return True

        connected_item: Presentation[UML.Element] | None
        connected_item = self.get_connected(self.element.handles()[0])
        assert connected_item
        Connector(connected_item, self.line).connect(handle, None)

        self.line.change_parent(self.element)

        return True

    def disconnect(self, handle):
        del self.line.subject

        for cinfo in self.diagram.connections.get_connections(connected=self.line):
            Connector(self.line, cinfo.item).disconnect(cinfo.handle)



"""Classes related (dependency, implementation) adapter connections."""

from __future__ import annotations

from gaphor import UML
from gaphor.core.modeling import Presentation
from gaphor.diagram.connectors import (
    Connector,
    DirectionalRelationshipConnect,
    RelationshipConnect,
)
from gaphor.diagram.presentation import Classified, Named
from gaphor.UML.classes.association import AssociationItem
from gaphor.UML.classes.dependency import DependencyItem
from gaphor.UML.recipes import owner_package


@Connector.register(Named, DependencyItem)
class DependencyConnect(DirectionalRelationshipConnect):
    """Connect two Named elements using a Dependency."""

    line: DependencyItem

    def connect_subject(self, handle):
        dependency_type = self.update_dependency_type(handle)

        relation = self.relationship_or_new(
            dependency_type,
            dependency_type.supplier,
            dependency_type.client,
        )
        self.line.subject = relation

        if not relation.owningPackage:
            assert isinstance(self.diagram, UML.Diagram)

            relation.owningPackage = owner_package(self.diagram.owner)

    def update_dependency_type(self, handle):
        line = self.line

        if line.auto_dependency:
            opposite = line.opposite(handle)

            other = self.get_connected(opposite)
            assert other
            if handle is line.head:
                client = other.subject
                supplier = self.element.subject
            else:
                client = self.element.subject
                supplier = other.subject
            line.dependency_type = UML.recipes.dependency_type(client, supplier)

        return line.dependency_type


@Connector.register(Classified, AssociationItem)
class AssociationConnect(RelationshipConnect):
    """Connect association to classifier."""

    line: AssociationItem

    def __init__(
        self, element: Presentation[UML.Element], line: AssociationItem
    ) -> None:
        super().__init__(element, line)
        self._navigabilities = (
            list(line.subject.memberEnd[:].navigability) if line.subject else []
        )

    def allow(self, handle, port):
        # Element should be a Classifier
        return super().allow(handle, port) and isinstance(
            self.element.subject, UML.Classifier
        )

    def relationship(  # type:ignore[override]
        self, head_subject, tail_subject
    ) -> UML.Association | None:
        line = self.line
        subject = line.subject

        def member_ends_match(subject):
            return len(subject.memberEnd) >= 2 and (
                (
                    head_subject is subject.memberEnd[0].type
                    and tail_subject is subject.memberEnd[1].type
                )
                or (
                    head_subject is subject.memberEnd[1].type
                    and tail_subject is subject.memberEnd[0].type
                )
            )

        # First check if the right subject is already connected:
        if line.subject and member_ends_match(line.subject):
            return subject

        diagram = self.diagram

        return next(
            (
                a
                for a in line.model.select(UML.Association)
                if member_ends_match(a) and diagram not in a.presentation[:].diagram
            ),
            None,
        )

    def new_relation(self, head_subject, tail_subject) -> UML.Association:  # type: ignore[override]
        return UML.recipes.create_association(head_subject, tail_subject)  # type: ignore[no-any-return]

    def relationship_or_new(self, head_subject, tail_subject) -> UML.Association:  # type: ignore[override]
        line = self.line

        relation: UML.Association | None = self.relationship(head_subject, tail_subject)

        if relation:
            return relation

        relation = self.new_relation_from_copy(UML.Association)

        if relation:
            relation.memberEnd[0].type = head_subject
            relation.memberEnd[1].type = tail_subject
            for end, nav in zip(relation.memberEnd, self._navigabilities, strict=False):
                UML.recipes.set_navigability(relation, end, nav)
        else:
            relation = self.new_relation(head_subject, tail_subject)
            tail_subject = relation.memberEnd[1]
            if (
                line.preferred_aggregation in ("shared", "composite")
                or line.preferred_tail_navigability == "navigable"
            ):
                UML.recipes.set_navigability(relation, tail_subject, True)
            tail_subject.aggregation = line.preferred_aggregation
            line.preferred_aggregation = "none"
            line.preferred_tail_navigability = "none"

        assert isinstance(relation, UML.Association)

        if not relation.package:
            assert isinstance(self.diagram, UML.Diagram)

            relation.package = owner_package(self.diagram.owner)

        return relation

    def connect_subject(self, handle):
        line = self.line
        head_end = self.get_connected(line.head)
        tail_end = self.get_connected(line.tail)
        assert head_end and tail_end
        relation = self.relationship_or_new(head_end.subject, tail_end.subject)

        if head_end.subject is relation.memberEnd[0].type:
            line.head_subject = relation.memberEnd[0]
            line.tail_subject = relation.memberEnd[1]
        else:
            line.head_subject = relation.memberEnd[1]
            line.tail_subject = relation.memberEnd[0]

        # Set subject last, to trigger updates on property editor, a.o.
        line.subject = relation

    def disconnect_subject(self, handle) -> None:
        del self.line.head_subject
        del self.line.tail_subject
        super().disconnect_subject(handle)



"""
Install our schema:

    mkdir -p ~/.local/share/glib-2.0/schemas
    cp data/org.gaphor.Gaphor.gschema.xml ~/.local/share/glib-2.0/schemas/
    glib-compile-schemas ~/.local/share/glib-2.0/schemas/

"""

import argparse
import shutil
import subprocess
from pathlib import Path

from gi.repository import GLib


def install_schemas_parser():
    parser = argparse.ArgumentParser(description="Install GSettings schema for Gaphor.")

    default_schema_dir = Path(GLib.get_user_data_dir()) / "glib-2.0" / "schemas"
    parser.add_argument(
        "--schemas-dir",
        dest="schemas_dir",
        metavar="path",
        default=default_schema_dir,
        help="installation directory for GSettings schemas",
    )
    parser.add_argument(
        "--glib-compile-schemas",
        dest="glib_compile_schemas",
        metavar="path",
        default="glib-compile-schemas",
        help="path to glib-compile-schemas",
    )
    parser.set_defaults(command=install_schemas_command)

    return parser


def install_schemas_command(args):
    schemas_dir = Path(args.schemas_dir)
    # Print, since logger is not initialized at this point
    print(f"Installing Gaphor schema in {schemas_dir}…")  # noqa: T201
    schemas_dir.mkdir(parents=True, exist_ok=True)
    schema_file = Path(__file__).parent / "org.gaphor.Gaphor.gschema.xml"
    shutil.copy(schema_file, schemas_dir)
    subprocess.run([args.glib_compile_schemas, str(schemas_dir)], check=True)



"""The action definition for the SysML toolbox."""

from gaphor import UML
from gaphor.diagram.diagramtoolbox import (
    DiagramType,
    DiagramTypes,
    ElementCreateInfo,
    ToolboxDefinition,
    ToolDef,
    ToolSection,
    new_item_factory,
)
from gaphor.i18n import gettext, i18nize
from gaphor.SysML import diagramitems as sysml_items
from gaphor.SysML.allocations.allocationstoolbox import allocations
from gaphor.SysML.blocks.blockstoolbox import blocks
from gaphor.SysML.diagramtype import DiagramDefault, SysMLDiagramType
from gaphor.SysML.requirements.requirementstoolbox import requirements
from gaphor.SysML.sysml import (
    ActivityDiagram,
    Block,
    BlockDefinitionDiagram,
    ConstraintBlock,
    InterfaceBlock,
    InternalBlockDiagram,
    PackageDiagram,
    Requirement,
    RequirementDiagram,
    SequenceDiagram,
    StateMachineDiagram,
    UseCaseDiagram,
    ValueType,
)
from gaphor.UML import diagramitems as uml_items
from gaphor.UML.actions.actionstoolbox import actions
from gaphor.UML.general.generaltoolbox import general_tools
from gaphor.UML.interactions.interactionstoolbox import interactions
from gaphor.UML.states.statestoolbox import states
from gaphor.UML.toolboxconfig import named_element_config
from gaphor.UML.uml import (
    Activity,
    Actor,
    Enumeration,
    Interaction,
    Package,
    StateMachine,
    UseCase,
)
from gaphor.UML.usecases.usecasetoolbox import use_cases

internal_blocks = ToolSection(
    gettext("Internal Blocks"),
    (
        ToolDef(
            "toolbox-connector",
            gettext("Connector"),
            "gaphor-connector-symbolic",
            "<Shift>C",
            new_item_factory(uml_items.ConnectorItem),
        ),
        ToolDef(
            "toolbox-property",
            gettext("Property"),
            "gaphor-property-symbolic",
            "o",
            new_item_factory(
                sysml_items.PropertyItem, UML.Property, config_func=named_element_config
            ),
        ),
        ToolDef(
            "toolbox-proxy-port",
            gettext("Proxy Port"),
            "gaphor-proxy-port-symbolic",
            "x",
            new_item_factory(sysml_items.ProxyPortItem),
        ),
    ),
)


sysml_toolbox_actions: ToolboxDefinition = (
    general_tools,
    blocks,
    internal_blocks,
    requirements,
    actions,
    interactions,
    states,
    use_cases,
    allocations,
)

root = type(None)

# Not implemented: Parameter Diagram
sysml_diagram_types: DiagramTypes = (
    SysMLDiagramType(
        BlockDefinitionDiagram,
        i18nize("Block Definition Diagram"),
        (blocks,),
        (Block, Package, ConstraintBlock, Activity),
        (DiagramDefault(root, Package, i18nize("New Package")),),
    ),
    SysMLDiagramType(
        InternalBlockDiagram,
        i18nize("Internal Block Diagram"),
        (internal_blocks,),
        (
            Block,
            ConstraintBlock,
        ),
        (
            DiagramDefault(Package, Block, i18nize("New Block")),
            DiagramDefault(root, Block, i18nize("New Block")),
        ),
    ),
    SysMLDiagramType(
        PackageDiagram,
        i18nize("Package Diagram"),
        (blocks,),
        (Package,),  # model, modelLibrary, profile
        (DiagramDefault(root, Package, i18nize("New Package")),),
    ),
    SysMLDiagramType(
        RequirementDiagram,
        i18nize("Requirement Diagram"),
        (requirements,),
        (
            Package,
            Requirement,
            # model,
            # modelLibrary
        ),
        (DiagramDefault(root, Package, i18nize("New Package")),),
    ),
    SysMLDiagramType(
        ActivityDiagram,
        i18nize("Activity Diagram"),
        (actions,),
        (Activity,),
        (
            DiagramDefault(Package, Activity, i18nize("New Activity")),
            DiagramDefault(root, Activity, i18nize("New Activity")),
        ),
    ),
    SysMLDiagramType(
        SequenceDiagram,
        i18nize("Sequence Diagram"),
        (interactions,),
        (Interaction,),
        (
            DiagramDefault(Package, Interaction, i18nize("New Interaction")),
            DiagramDefault(root, Interaction, i18nize("New Interaction")),
        ),
    ),
    SysMLDiagramType(
        StateMachineDiagram,
        i18nize("State Machine Diagram"),
        (states,),
        (StateMachine,),
        (
            DiagramDefault(Package, StateMachine, i18nize("New State Machine")),
            DiagramDefault(root, StateMachine, i18nize("New State Machine")),
        ),
    ),
    SysMLDiagramType(
        UseCaseDiagram,
        i18nize("Use Case Diagram"),
        (use_cases,),
        (
            Package,
            Block,
            # model,
            # modelLibrary
        ),
        (DiagramDefault(root, Package, i18nize("New Package")),),
    ),
    DiagramType(UML.Diagram, i18nize("Generic Diagram"), ()),
)

sysml_element_types = (
    ElementCreateInfo("package", i18nize("Package"), Package, (Package,)),
    ElementCreateInfo("activity", i18nize("Activity"), Activity, (Package,)),
    ElementCreateInfo("actor", i18nize("Actor"), Actor, (Package,)),
    ElementCreateInfo("block", i18nize("Block"), Block, (Package,)),
    ElementCreateInfo("enumeration", i18nize("Enumeration"), Enumeration, (Package,)),
    ElementCreateInfo("interaction", i18nize("Interaction"), Interaction, (Package,)),
    ElementCreateInfo(
        "interfaceblock", i18nize("Interface Block"), InterfaceBlock, (Package,)
    ),
    ElementCreateInfo(
        "requirement",
        i18nize("Requirement"),
        Requirement,
        (
            Package,
            Requirement,
        ),
    ),
    ElementCreateInfo(
        "statemachine", i18nize("State Machine"), StateMachine, (Package,)
    ),
    ElementCreateInfo("usecase", i18nize("Use Case"), UseCase, (Package,)),
    ElementCreateInfo("valuetype", i18nize("Value Type"), ValueType, (Package,)),
)



from __future__ import annotations

from gaphor import UML
from gaphor.diagram.propertypages import (
    PropertyPageBase,
    PropertyPages,
    handler_blocking,
    new_resource_builder,
    unsubscribe_all_on_destroy,
)
from gaphor.SysML import sysml
from gaphor.SysML.blocks.block import BlockItem
from gaphor.SysML.blocks.interfaceblock import InterfaceBlockItem
from gaphor.SysML.blocks.property import PropertyItem
from gaphor.SysML.blocks.proxyport import ProxyPortItem
from gaphor.SysML.requirements import RequirementItem
from gaphor.transaction import Transaction
from gaphor.UML.classes.classespropertypages import (
    ATTRIBUTES_PAGE_CLASSIFIERS,
    OPERATIONS_PAGE_CLASSIFIERS,
    ShowOperationsPage,
)
from gaphor.UML.propertypages import (
    ShowTypedElementPropertyPage,
    TypedElementPropertyPage,
    list_of_classifiers,
)

new_builder = new_resource_builder("gaphor.SysML")


@PropertyPages.register(sysml.Requirement)
class RequirementPropertyPage(PropertyPageBase):
    order = 15

    def __init__(self, subject: sysml.Requirement, event_manager):
        super().__init__()
        assert subject
        self.subject = subject
        self.event_manager = event_manager
        self.watcher = subject.watcher()

    def construct(self):
        builder = new_builder(
            "requirement-editor",
            "requirement-text-buffer",
            signals={
                "requirement-id-changed": (self._on_id_changed,),
            },
        )
        subject = self.subject

        entry = builder.get_object("requirement-id")
        entry.set_text(subject.externalId or "")

        text_view = builder.get_object("requirement-text")

        buffer = builder.get_object("requirement-text-buffer")
        if subject.text:
            buffer.set_text(subject.text)

        @handler_blocking(buffer, "changed", self._on_text_changed)
        def text_handler(event):
            if not text_view.props.has_focus:
                buffer.set_text(event.new_value)

        self.watcher.watch("text", text_handler)

        return unsubscribe_all_on_destroy(
            builder.get_object("requirement-editor"), self.watcher
        )

    def _on_id_changed(self, entry):
        with Transaction(self.event_manager, context="editing"):
            self.subject.externalId = entry.get_text()

    def _on_text_changed(self, buffer):
        with Transaction(self.event_manager, context="editing"):
            self.subject.text = buffer.get_text(
                buffer.get_start_iter(), buffer.get_end_iter(), False
            )


@PropertyPages.register(RequirementItem)
class RequirementItemPropertyPage(PropertyPageBase):
    order = 16

    def __init__(self, item: RequirementItem, event_manager):
        super().__init__()
        self.item = item
        self.event_manager = event_manager

    def construct(self):
        if not self.item.subject:
            return

        builder = new_builder(
            "requirement-item-editor",
            signals={
                "show-requirement-text-changed": (self._on_show_text_change,),
            },
        )

        show_requirement_text = builder.get_object("show-requirement-text")
        show_requirement_text.set_active(self.item.show_text)

        return builder.get_object("requirement-item-editor")

    def _on_show_text_change(self, button, gparam):
        with Transaction(self.event_manager):
            self.item.show_text = button.get_active()


ATTRIBUTES_PAGE_CLASSIFIERS.extend([sysml.ValueType])
OPERATIONS_PAGE_CLASSIFIERS.extend([sysml.Block, sysml.InterfaceBlock, sysml.ValueType])
PropertyPages.register(BlockItem)(ShowOperationsPage)
PropertyPages.register(InterfaceBlockItem)(ShowOperationsPage)

PropertyPages.register(UML.Property)(TypedElementPropertyPage)
PropertyPages.register(PropertyItem)(ShowTypedElementPropertyPage)
PropertyPages.register(sysml.ProxyPort)(TypedElementPropertyPage)
PropertyPages.register(ProxyPortItem)(ShowTypedElementPropertyPage)


@PropertyPages.register(BlockItem)
class CompartmentPage(PropertyPageBase):
    """An editor for Block items."""

    order = 30

    def __init__(self, item, event_manager):
        super().__init__()
        self.item = item
        self.event_manager = event_manager

    def construct(self):
        if not self.item.subject:
            return

        builder = new_builder(
            "compartment-editor",
            signals={
                "show-parts-changed": (self._on_show_parts_change,),
                "show-references-changed": (self._on_show_references_change,),
                "show-values-changed": (self._on_show_values_change,),
            },
        )

        show_parts = builder.get_object("show-parts")
        show_parts.set_active(self.item.show_parts)

        show_references = builder.get_object("show-references")
        show_references.set_active(self.item.show_references)

        show_values = builder.get_object("show-values")
        show_values.set_active(self.item.show_values)

        return builder.get_object("compartment-editor")

    def _on_show_parts_change(self, button, gparam):
        with Transaction(self.event_manager):
            self.item.show_parts = button.get_active()

    def _on_show_references_change(self, button, gparam):
        with Transaction(self.event_manager):
            self.item.show_references = button.get_active()

    def _on_show_values_change(self, button, gparam):
        with Transaction(self.event_manager):
            self.item.show_values = button.get_active()


@PropertyPages.register(InterfaceBlockItem)
class InterfaceBlockPage(PropertyPageBase):
    """An editor for InterfaceBlock items."""

    order = 30

    def __init__(self, item, event_manager):
        super().__init__()
        self.item = item
        self.event_manager = event_manager

    def construct(self):
        if not self.item.subject:
            return

        builder = new_builder(
            "interfaceblock-editor",
            signals={
                "show-values-changed": (self._on_show_values_change,),
            },
        )

        show_values = builder.get_object("show-values")
        show_values.set_active(self.item.show_values)

        return builder.get_object("interfaceblock-editor")

    def _on_show_values_change(self, button, gparam):
        with Transaction(self.event_manager):
            self.item.show_values = button.get_active()


@PropertyPages.register(sysml.Property)
class PropertyAggregationPropertyPage(PropertyPageBase):
    """An editor for Properties (a.k.a.

    attributes).
    """

    order = 30

    AGGREGATION = ("none", "shared", "composite")

    def __init__(self, subject: sysml.Property, event_manager):
        super().__init__()
        self.subject = subject
        self.event_manager = event_manager

    def construct(self):
        if not self.subject or isinstance(self.subject, UML.Port):
            return

        builder = new_builder(
            "property-aggregation-editor",
            signals={
                "aggregation-changed": (self._on_aggregation_change,),
            },
        )

        aggregation = builder.get_object("aggregation")
        aggregation.set_selected(self.AGGREGATION.index(self.subject.aggregation))

        return builder.get_object("property-aggregation-editor")

    def _on_aggregation_change(self, combo, _pspec):
        with Transaction(self.event_manager):
            self.subject.aggregation = self.AGGREGATION[combo.get_selected()]


@PropertyPages.register(UML.Association)
@PropertyPages.register(sysml.Connector)
class ItemFlowPropertyPage(PropertyPageBase):
    """Item Flow on Connectors."""

    order = 35

    def __init__(self, subject: UML.Association | sysml.Connector, event_manager):
        super().__init__()
        self.subject = subject
        self.event_manager = event_manager

    @property
    def information_flow(self):
        subject = self.subject
        if isinstance(subject, sysml.Connector):
            iflows = subject.informationFlow
        elif isinstance(subject, UML.Relationship):
            iflows = subject.abstraction
        return iflows[0] if iflows else None

    def construct(self):
        if not self.subject:
            return

        builder = new_builder(
            "item-flow-editor",
            signals={
                "item-flow-active": (self._on_item_flow_active,),
                "item-flow-name-changed": (self._on_item_flow_name_changed,),
                "invert-direction-changed": (self.on_invert_direction_changed,),
            },
        )

        use_flow = builder.get_object("use-item-flow")
        self.entry = builder.get_object("item-flow-name")

        dropdown = builder.get_object("item-flow-type")
        model = list_of_classifiers(self.subject.model, UML.Classifier)
        dropdown.set_model(model)

        use_flow.set_active(isinstance(self.information_flow, sysml.ItemFlow))
        use_flow.set_sensitive(
            not (
                self.information_flow
                and type(self.information_flow) is UML.InformationFlow
            )
        )
        self.entry.set_sensitive(use_flow.get_active())
        if (iflow := self.information_flow) and iflow.itemProperty:
            assert isinstance(iflow, sysml.ItemFlow)
            self.entry.set_text(iflow.itemProperty.name or "")
            if iflow.itemProperty.type:
                dropdown.set_selected(
                    next(
                        n
                        for n, lv in enumerate(model)
                        if lv.value == iflow.itemProperty.type.id
                    )
                )

        dropdown.connect("notify::selected", self._on_item_flow_type_changed)

        return builder.get_object("item-flow-editor")

    def _on_item_flow_active(self, switch, gparam):
        active = switch.get_active()
        subject = self.subject
        iflow = self.information_flow
        with Transaction(self.event_manager):
            if active and not iflow:
                if isinstance(subject, sysml.Connector):
                    subject.informationFlow = create_item_flow(subject)
                elif isinstance(subject, UML.Relationship):
                    subject.abstraction = create_item_flow(subject)
            elif not active and iflow:
                iflow.unlink()
        self.entry.set_sensitive(switch.get_active())
        self.entry.set_text("")

    def _on_item_flow_name_changed(self, entry):
        if not (iflow := self.information_flow):
            return

        assert isinstance(iflow, sysml.ItemFlow)
        with Transaction(self.event_manager, context="editing"):
            iflow.itemProperty.name = entry.get_text()

    def _on_item_flow_type_changed(self, dropdown, _pspec):
        if not (iflow := self.information_flow):
            return

        assert isinstance(iflow, sysml.ItemFlow)
        with Transaction(self.event_manager):
            if id := dropdown.get_selected_item().value:
                element = self.subject.model.lookup(id)
                assert isinstance(element, UML.Type)
                iflow.itemProperty.type = element
            else:
                del iflow.itemProperty.type

    def on_invert_direction_changed(self, button):
        if not (iflow := self.information_flow):
            return

        with Transaction(self.event_manager):
            iflow.informationSource, iflow.informationTarget = (
                iflow.informationTarget,
                iflow.informationSource,
            )


def create_item_flow(subject: UML.Association | sysml.Connector) -> sysml.ItemFlow:
    iflow = subject.model.create(sysml.ItemFlow)
    if isinstance(subject, sysml.Connector):
        iflow.informationSource = subject.end[0].role
        iflow.informationTarget = subject.end[1].role
    elif isinstance(subject, UML.Relationship):
        iflow.informationSource = subject.memberEnd[0]
        iflow.informationTarget = subject.memberEnd[1]
    iflow.itemProperty = subject.model.create(sysml.Property)
    return iflow



class ToolCompleted:
    pass


class DiagramItemPlaced(ToolCompleted):
    def __init__(self, item):
        self.item = item


class DiagramOpened:
    def __init__(self, diagram):
        self.diagram = diagram


class DiagramClosed:
    def __init__(self, diagram):
        self.diagram = diagram


class DiagramSelectionChanged:
    def __init__(self, diagram_view, focused_item, selected_items):
        self.diagram_view = diagram_view
        self.focused_item = focused_item
        self.selected_items = selected_items



"""A generic way to handle errors in GUI applications.

This module also contains a ErrorHandlerAspect, which can be easily
attached to a class method and will raise the error dialog when the
method exits with an exception.
"""

import pdb
import sys

from gi.repository import Adw

from gaphor.i18n import gettext


async def error_dialog(message, secondary_message="", window=None):
    _exc_type, _exc_value, exc_traceback = sys.exc_info()
    atty = __debug__ and exc_traceback and sys.stdin.isatty()

    debug_body = (f"{secondary_message}\n\n" if secondary_message else "") + gettext(
        "It looks like Gaphor is started from the command line. Do you want to open a debug session?"
    )
    dialog = Adw.AlertDialog.new(
        message,
        debug_body if atty else secondary_message,
    )
    dialog.add_response("close", gettext("Close"))
    if atty:
        dialog.add_response("debug", gettext("Start Debug Session"))
        dialog.set_default_response("debug")
    else:
        dialog.set_default_response("close")
    dialog.set_close_response("close")

    answer = await dialog.choose(window)

    if exc_traceback and answer in (100, "debug"):
        pdb.post_mortem(exc_traceback)



import asyncio
import base64
import io
from pathlib import Path

from PIL import Image

from gaphor.core import gettext
from gaphor.diagram.propertypages import PropertyPageBase, PropertyPages, new_builder
from gaphor.transaction import Transaction
from gaphor.ui.errordialog import error_dialog
from gaphor.ui.filedialog import open_file_dialog
from gaphor.UML.general.image import ImageItem


@PropertyPages.register(ImageItem)
class ImagePropertyPage(PropertyPageBase):
    """Edit image settings"""

    def __init__(self, subject, event_manager):
        super().__init__()
        self.subject = subject
        self.event_manager = event_manager
        self.watcher = subject and subject.watcher()
        self.background_task = None

    def construct(self):
        subject = self.subject

        if not subject:
            return

        builder = new_builder(
            "picture-editor",
            signals={
                "select-picture": (self._on_select_picture_clicked,),
                "set-default-size": (self._on_default_size_clicked),
            },
        )
        return builder.get_object("picture-editor")

    async def open_file(self, parent):
        filename: Path = await open_file_dialog(  # type: ignore[assignment]
            gettext("Select a picture..."),
            image_filter=True,
            parent=parent,
            multiple=False,
        )
        if not filename:
            return

        with open(filename, "rb") as file:
            try:
                image_data = file.read()
                image = Image.open(io.BytesIO(image_data))
                image.verify()
                image.close()

                base64_encoded_data = base64.b64encode(image_data)

                with Transaction(self.event_manager):
                    self.subject.subject.content = base64_encoded_data.decode("ascii")
                    self.subject.width = image.width
                    self.subject.height = image.height
                    if self.subject.subject.name in [
                        None,
                        gettext("New Image"),
                    ] and (new_image_name := self.sanitize_image_name(filename)):
                        self.subject.subject.name = new_image_name
            except Exception:
                await error_dialog(
                    message=gettext("Unable to parse picture “{filename}”.").format(
                        filename=filename
                    )
                )

    def _on_select_picture_clicked(self, button):
        if self.background_task and not self.background_task.done():
            return

        self.background_task = asyncio.create_task(self.open_file(button.get_root()))

    def _on_default_size_clicked(self, button):
        if self.subject and self.subject.subject and self.subject.subject.content:
            base64_img_bytes = self.subject.subject.content.encode("ascii")
            image_data = base64.decodebytes(base64_img_bytes)
            image = Image.open(io.BytesIO(image_data))

            with Transaction(self.event_manager):
                self.subject.width = image.width
                self.subject.height = image.height

    def sanitize_image_name(self, filename):
        return "".join(
            chr if chr.isalnum() or (chr in [" ", "_", "-"]) else "_"
            for chr in Path(filename).stem
        )



import gaphor.diagram.align
from gaphor.abc import ActionProvider, Service
from gaphor.action import action
from gaphor.core import event_handler
from gaphor.diagram.event import DiagramSelectionChanged
from gaphor.diagram.presentation import ElementPresentation
from gaphor.event import ActionEnabled
from gaphor.transaction import Transaction


class DiagramAlign(Service, ActionProvider):
    def __init__(self, event_manager, diagrams):
        self.event_manager = event_manager
        self.diagrams = diagrams

        event_manager.subscribe(self._on_selection_changed)

    def shutdown(self):
        self.event_manager.unsubscribe(self._on_selection_changed)

    @property
    def alignable_elements(self):
        return (
            {
                item
                for item in view.selection.selected_items
                if isinstance(item, ElementPresentation)
            }
            if (view := self.diagrams.get_current_view())
            else ()
        )

    @action(name="win.diagram-align")
    def align_elements(self, target: str):
        if (view := self.diagrams.get_current_view()) and (
            len(elements := self.alignable_elements) > 1
        ):
            with Transaction(self.event_manager):
                align[target](elements)
                view.model.update(view.model.get_all_items())

    @event_handler(DiagramSelectionChanged)
    def _on_selection_changed(self, event=None):
        self.event_manager.handle(
            ActionEnabled("win.diagram-align", len(self.alignable_elements) > 1)
        )


align = {
    "left": gaphor.diagram.align.align_left,
    "right": gaphor.diagram.align.align_right,
    "vertical-center": gaphor.diagram.align.align_vertical_center,
    "top": gaphor.diagram.align.align_top,
    "bottom": gaphor.diagram.align.align_bottom,
    "horizontal-center": gaphor.diagram.align.align_horizontal_center,
    "max-height": gaphor.diagram.align.resize_max_height,
    "max-width": gaphor.diagram.align.resize_max_width,
    "max-size": gaphor.diagram.align.resize_max_size,
}



from gaphor.UML.profiles import extensionconnect  # noqa: F401
from gaphor.UML.profiles.extension import ExtensionItem
from gaphor.UML.profiles.packageimport import PackageImportItem

__all__ = ["ExtensionItem", "PackageImportItem"]



"""Transaction support for Gaphor."""

from __future__ import annotations

import logging

from gaphor.event import TransactionBegin, TransactionCommit, TransactionRollback

log = logging.getLogger(__name__)


class TransactionError(Exception):
    """Errors related to the transaction module."""


class Transaction:
    """The transaction.

    On start and end of a transaction an event is emitted.
    Transactions can be nested. Events are only emitted when the
    outermost transaction begins and finishes.

    Note that transactions are a global construct.

    >>> import gaphor.core.eventmanager
    >>> event_manager = gaphor.core.eventmanager.EventManager()

    Transactions can be nested. If the outermost transaction is committed or
    rolled back, an event is emitted.

    It's most convenient to use ``Transaction`` as a context manager:

    >>> with Transaction(event_manager) as ctx:
    ...     ... # do actions
    ...     # in case the transaction should be rolled back:
    ...     ctx.rollback()

    Events can be handled programmatically, although this is discouraged:

    >>> tx = Transaction(event_manager)
    >>> tx.commit()

    """

    _stack: list[Transaction] = []

    def __init__(self, event_manager, context: str | None = None):
        """Initialize the transaction.

        If this is the first transaction in the stack, a
        :obj:`~gaphor.event.TransactionBegin` event is emitted.
        """
        self.event_manager = event_manager
        self.context = context

        self._need_rollback = False
        if not self._stack:
            self._handle(TransactionBegin(self.context))
        self._stack.append(self)

    def commit(self):
        """Commit the transaction.

        The transaction is closed. A :obj:`~gaphor.event.TransactionCommit` event is emitted.
        If the transaction needs to be rolled back,
        a :obj:`~gaphor.event.TransactionRollback` event is emitted instead.
        """

        self._close()
        if not self._stack:
            if self._need_rollback:
                self._handle(TransactionRollback(self.context))
            else:
                self._handle(TransactionCommit(self.context))

    def rollback(self):
        """Roll-back the transaction.

        First, the transaction is closed.
        A :obj:`~gaphor.event.TransactionRollback` event is emitted.
        """

        self.mark_rollback()
        self.commit()

    @classmethod
    def mark_rollback(cls):
        """Mark the transaction for rollback.

        This operation itself will not close the transaction,
        instead it will allow you to elegantly revert changes.
        """
        for tx in cls._stack:
            tx._need_rollback = True  # noqa: SLF001

    @classmethod
    def in_transaction(cls) -> bool:
        """Are you running inside a transaction?"""
        return bool(cls._stack)

    def _close(self):
        try:
            last = self._stack.pop()
        except IndexError:
            raise TransactionError("No Transaction on stack.") from None
        if last is not self:
            self._stack.append(last)
            raise TransactionError(
                "Transaction on stack is not the transaction being closed."
            )

    def _handle(self, event):
        self.event_manager.handle(event)

    def __enter__(self) -> TransactionContext:
        """Provide ``with``-statement transaction support."""
        return TransactionContext(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Provide ``with``-statement transaction support.

        If an error occurred, the transaction is rolled back. Otherwise,
        it is committed.
        """

        if exc_type and not self._need_rollback:
            log.error(
                "Transaction terminated due to an exception, performing a rollback",
            )
            self.mark_rollback()
        self.commit()


class TransactionContext:
    """A simple context for a transaction.

    Can only perform a rollback.
    """

    def __init__(self, tx: Transaction) -> None:
        self._tx = tx

    def rollback(self) -> None:
        self._tx.mark_rollback()



"""The lexical analyzer for attributes and operations.

In this module some parse functions are added for attributes and
operations. The regular expressions are constructed based on a series of
"sub-patterns". This makes it easy to identify the autonomy of an
attribute/operation.
"""

import contextlib

__all__ = ["parse_property", "parse_operation"]

import re

from gaphor.core.format import parse
from gaphor.UML import recipes, uml

# Visibility (optional) ::= '+' | '-' | '#'
vis_subpat = r"\s*(?P<vis>[-+#])?"

# Derived value (optional) ::= [/]
derived_subpat = r"\s*(?P<derived>/)?"

# name (required) ::= name
name_subpat = r"\s*(?P<name>[a-zA-Z_]\w*( +\w+)*)"

# Multiplicity ::= '[' [mult_l ..] mult_u ']'
mult_subpat = r"\s*(?P<has_mult>\[\s*((?P<mult_l>[0-9]+)\s*\.\.)?\s*(?P<mult_u>([0-9]+|\*))?\s*\])?"
multa_subpat = r"\s*(\[?((?P<mult_l>[0-9]+)\s*\.\.)?\s*(?P<mult_u>([0-9]+|\*))\]?)?"

# Type and multiplicity (optional) ::= ':' type
type_subpat = r"\s*(:\s*(?P<type>[a-zA-Z_]\w*( +\w+| *\| *\w+| *<[\w\| ]*>)*))?"

# default value (optional) ::= '=' default
default_subpat = (
    r"\s*(=\s*(?P<default>((?P<oq>[\"'])(?:(?=(?P<ec>\\?))(?P=ec).)*?(?P=oq))|(\S+)))?"
)

# tagged values (optional) ::= '{' tags '}'
tags_subpat = r"\s*(\{\s*(?P<tags>.*?)\s*\})?"

# Parameters (required) ::= '(' [params] ')'
params_subpat = r"\s*\(\s*(?P<params>[^)]+)?\)"

# Possible other parameters (optional) ::= ',' rest
rest_subpat = r"\s*(,\s*(?P<rest>.*))?"

# Direction of a parameter (optional, default in) ::= 'in' | 'out' | 'inout'
dir_subpat = r"\s*((?P<dir>in|out|inout)\s)?"

# A note ::= '#' text
note_subpat = r"\s*(#\s*(?P<note>.*))?"

# Some trailing garbage => no valid syntax...
garbage_subpat = r"\s*(?P<garbage>.*)"


def compile(regex):
    return re.compile(regex, re.MULTILINE | re.DOTALL)


# Attribute:
#   [+-#] [/] name [: type[\[mult\]]] [= default] [{ tagged values }]
attribute_pat = compile(
    r"^"
    + vis_subpat
    + derived_subpat
    + name_subpat
    + type_subpat
    + mult_subpat
    + default_subpat
    + tags_subpat
    + note_subpat
    + garbage_subpat
)

# Association end name:
#   [[+-#] [/] name [\[mult\]]] [{ tagged values }]
association_end_name_pat = compile(
    r"^"
    + "("
    + vis_subpat
    + derived_subpat
    + name_subpat
    + type_subpat
    + mult_subpat
    + ")?"
    + tags_subpat
    + note_subpat
    + garbage_subpat
)

# Association end multiplicity:
#   [mult] [{ tagged values }]
association_end_mult_pat = compile(
    f"^{vis_subpat}{derived_subpat}{multa_subpat}{tags_subpat}{garbage_subpat}"
)


# Operation:
#   [+|-|#] name ([parameters]) [: type[\[mult\]]] [{ tagged values }]
operation_pat = compile(
    r"^"
    + vis_subpat
    + name_subpat
    + params_subpat
    + type_subpat
    + mult_subpat
    + tags_subpat
    + note_subpat
    + garbage_subpat
)

# One parameter supplied with an operation:
#   [in|out|inout] name [: type[\[mult\]] [{ tagged values }]
parameter_pat = compile(
    r"^"
    + dir_subpat
    + name_subpat
    + type_subpat
    + mult_subpat
    + default_subpat
    + note_subpat
    + garbage_subpat
)

parameters_pat = compile(
    r"^"
    + dir_subpat
    + name_subpat
    + type_subpat
    + mult_subpat
    + default_subpat
    + tags_subpat
    + rest_subpat
)

# Lifeline:
#  [name] [: type]  # noqa: E800
lifeline_pat = compile(f"^{name_subpat}{type_subpat}{mult_subpat}{garbage_subpat}")


def _set_visibility(el: uml.Feature, vis: str):
    if vis == "#":
        el.visibility = "protected"
    elif vis == "+":
        el.visibility = "public"
    elif vis == "-":
        el.visibility = "private"
    elif vis == "~":
        el.visibility = "package"
    else:
        with contextlib.suppress(AttributeError):
            del el.visibility


def parse_attribute(el: uml.Property, s: str) -> None:
    """Parse string s in the property.

    Tagged values, multiplicity and stuff like that is altered to
    reflect the data in the property string.
    """
    m = attribute_pat.match(s)
    if not m or m.group("garbage"):
        el.name = s
        del el.visibility
        del el.isDerived
        if el.typeValue:
            el.typeValue = None
        if el.lowerValue:
            el.lowerValue = None
        if el.upperValue:
            el.upperValue = None
        if el.defaultValue:
            el.defaultValue = None
        if el.note:
            el.note = None
    else:
        g = m.group
        _set_visibility(el, g("vis"))
        el.isDerived = bool(g("derived"))
        el.name = g("name")
        el.typeValue = g("type")
        recipes.set_multiplicity_lower_value(el, g("mult_l"))
        recipes.set_multiplicity_upper_value(el, g("mult_u"))
        if g("has_mult") and not g("mult_u"):
            recipes.set_multiplicity_upper_value(el, "*")
        recipes.set_default_value_from_string(el, g("default"))
        el.note = g("note")


def parse_association_end(el: uml.Property, s: str) -> None:
    """Parse the text at one end of an association.

    The association end holds two strings. It is automatically figured
    out which string is fed to the parser.
    """
    # if no name, then clear as there could be some garbage
    # due to previous parsing (i.e. '[1'
    m = association_end_name_pat.match(s)
    if m and not m.group("name"):
        el.name = None

    # clear also multiplicity if no characters in ``s``
    m = association_end_mult_pat.match(s)
    if m and not m.group("mult_u") and el.upperValue:
        el.upperValue.value = None
        el.upperValue.name = None

    if m and m.group("mult_u") or m.group("tags"):
        g = m.group
        _set_visibility(el, g("vis"))
        el.isDerived = bool(g("derived"))
        recipes.set_multiplicity_lower_value(el, g("mult_l"))
        recipes.set_multiplicity_upper_value(el, g("mult_u"))
    else:
        m = association_end_name_pat.match(s)
        g = m.group
        if g("garbage"):
            el.name = s
            del el.visibility
            del el.isDerived
            del el.note
        else:
            _set_visibility(el, g("vis"))
            el.isDerived = bool(g("derived"))
            el.name = g("name")
            el.note = g("note")
            # Optionally, the multiplicity and tagged values may be defined:
            if g("mult_l"):
                recipes.set_multiplicity_lower_value(el, g("mult_l"))

            if g("mult_u"):
                if not g("mult_l"):
                    recipes.set_multiplicity_lower_value(el, None)
                recipes.set_multiplicity_upper_value(el, g("mult_u"))

            elif g("has_mult") and not g("mult_u"):
                recipes.set_multiplicity_upper_value(el, "*")


@parse.register(uml.Element)
def parse_namedelement(el: uml.Element, text: str) -> None:
    """Parse element by simply assigning text to its name."""
    el.name = text


@parse.register(uml.Property)
def parse_property(el: uml.Property, s: str) -> None:
    if el.association:
        parse_association_end(el, s)
    else:
        parse_attribute(el, s)


@parse.register(uml.Operation)
def parse_operation(el: uml.Operation, s: str) -> None:
    """Parse string s in the operation.

    Tagged values, parameters and visibility is altered to reflect the
    data in the operation string.
    """
    m = operation_pat.match(s)
    if not m or m.group("garbage"):
        el.name = s
        del el.visibility
        for param in list(el.ownedParameter):
            param.unlink()
    else:
        g = m.group
        create = el.model.create
        _set_visibility(el, g("vis"))
        el.name = g("name")
        el.note = g("note")

        defined_params = set()
        if g("type"):
            if returnParameters := [
                p for p in el.ownedParameter if p.direction == "return"
            ]:
                p = returnParameters[0]
            else:
                p = create(uml.Parameter)
                el.ownedParameter = p
                p.direction = "return"
            p.typeValue = g("type")
            recipes.set_multiplicity_lower_value(p, g("mult_l"))
            recipes.set_multiplicity_upper_value(p, g("mult_u"))
            if g("has_mult") and not g("mult_u"):
                recipes.set_multiplicity_upper_value(p, "*")
            defined_params.add(p)

        pindex = 0
        params = g("params")
        while params:
            m = parameters_pat.match(params)
            if not m:
                break
            g = m.group
            try:
                while el.ownedParameter[pindex] in defined_params:
                    pindex += 1
                p = el.ownedParameter[pindex]
            except IndexError:
                p = create(uml.Parameter)
            p.direction = g("dir") or "in"
            p.name = g("name")
            p.typeValue = g("type")
            recipes.set_multiplicity_lower_value(p, g("mult_l"))
            recipes.set_multiplicity_upper_value(p, g("mult_u"))
            if g("has_mult") and not g("mult_u"):
                recipes.set_multiplicity_upper_value(p, "*")
            recipes.set_default_value_from_string(p, g("default"))
            el.ownedParameter = p
            defined_params.add(p)
            # Do the next parameter:
            params = g("rest")

        # Remove remaining parameters:
        for op in list(el.ownedParameter):
            if op not in defined_params:
                op.unlink()


@parse.register(uml.Parameter)
def parse_parameter(el: uml.Parameter, s: str) -> None:
    m = parameter_pat.match(s)
    if not m or m.group("garbage"):
        el.name = s
        del el.direction
        del el.typeValue
    else:
        g = m.group
        el.direction = g("dir") or "in"
        el.name = g("name")
        el.typeValue = g("type")
        recipes.set_multiplicity_lower_value(el, g("mult_l"))
        recipes.set_multiplicity_upper_value(el, g("mult_u"))
        if g("has_mult") and not g("mult_u"):
            recipes.set_multiplicity_upper_value(el, "*")
        recipes.set_default_value_from_string(el, g("default"))


def parse_lifeline(el: uml.Lifeline, s: str) -> None:
    """Parse string s in a lifeline.

    If a class is defined and can be found in the datamodel, then a
    class is connected to the lifelines 'represents' property.
    """
    m = lifeline_pat.match(s)
    g = m.group
    if not m or g("garbage"):
        el.name = s
    else:
        el.name = g("name") + ": "
        if t := g("type"):
            el.name += f": {t}"
        # In the near future the data model should be extended with
        # Lifeline.represents: ConnectableElement  # noqa: E800


def render_lifeline(el: uml.Lifeline) -> str:
    return el.name or ""



import logging

from gaphor import UML
from gaphor.core.format import format
from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import (
    Classified,
    ElementPresentation,
)
from gaphor.diagram.shapes import Box, CssNode, Text, draw_border, draw_top_separator
from gaphor.diagram.support import represents
from gaphor.UML.classes.stereotype import stereotype_compartments, stereotype_watches
from gaphor.UML.compartments import name_compartment

log = logging.getLogger(__name__)


@represents(UML.Class)
@represents(UML.Stereotype)
class ClassItem(Classified, ElementPresentation[UML.Class]):
    """This item visualizes a Class instance.

    A ClassItem contains two compartments: one for attributes and one
    for operations.
    """

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id=id)
        self.watch("show_stereotypes", self.update_shapes).watch(
            "show_attributes", self.update_shapes
        ).watch("show_operations", self.update_shapes).watch(
            "subject[NamedElement].name"
        ).watch("subject[NamedElement].namespace.name").watch(
            "subject[Classifier].isAbstract", self.update_shapes
        )
        attribute_watches(self, "Class")
        operation_watches(self, "Class")
        stereotype_watches(self)

    show_stereotypes: attribute[int] = attribute("show_stereotypes", int)

    show_attributes: attribute[int] = attribute("show_attributes", int, default=True)

    show_operations: attribute[int] = attribute("show_operations", int, default=True)

    def additional_stereotypes(self):
        if isinstance(self.subject, UML.Stereotype):
            return [self.diagram.gettext("stereotype")]
        elif UML.recipes.is_metaclass(self.subject):
            return [self.diagram.gettext("metaclass")]
        return ()

    def update_shapes(self, event=None):
        self.shape = Box(
            name_compartment(self, self.additional_stereotypes),
            *(
                self.show_attributes
                and self.subject
                and [attributes_compartment(self.subject)]
                or []
            ),
            *(
                self.show_operations
                and self.subject
                and [operations_compartment(self.subject)]
                or []
            ),
            *(self.show_stereotypes and stereotype_compartments(self.subject) or []),
            draw=draw_border,
        )


def attribute_watches(presentation, cast):
    presentation.watch(
        f"subject[{cast}].ownedAttribute", presentation.update_shapes
    ).watch(
        f"subject[{cast}].ownedAttribute.association", presentation.update_shapes
    ).watch(f"subject[{cast}].ownedAttribute.name").watch(
        f"subject[{cast}].ownedAttribute.isStatic", presentation.update_shapes
    ).watch(
        f"subject[{cast}].ownedAttribute.isReadOnly", presentation.update_shapes
    ).watch(f"subject[{cast}].ownedAttribute.isDerived").watch(
        f"subject[{cast}].ownedAttribute.visibility"
    ).watch(f"subject[{cast}].ownedAttribute.lowerValue").watch(
        f"subject[{cast}].ownedAttribute.upperValue"
    ).watch(f"subject[{cast}].ownedAttribute.defaultValue").watch(
        f"subject[{cast}].ownedAttribute.type"
    ).watch(f"subject[{cast}].ownedAttribute.type.name").watch(
        f"subject[{cast}].ownedAttribute.typeValue"
    )


def operation_watches(presentation, cast):
    presentation.watch(
        f"subject[{cast}].ownedOperation", presentation.update_shapes
    ).watch(f"subject[{cast}].ownedOperation.name").watch(
        f"subject[{cast}].ownedOperation.isAbstract", presentation.update_shapes
    ).watch(
        f"subject[{cast}].ownedOperation.isStatic", presentation.update_shapes
    ).watch(f"subject[{cast}].ownedOperation.visibility").watch(
        f"subject[{cast}].ownedOperation.ownedParameter.lowerValue"
    ).watch(f"subject[{cast}].ownedOperation.ownedParameter.upperValue").watch(
        f"subject[{cast}].ownedOperation.ownedParameter.type.name"
    ).watch(f"subject[{cast}].ownedOperation.ownedParameter.typeValue").watch(
        f"subject[{cast}].ownedOperation.ownedParameter.defaultValue"
    )


def attributes_compartment(subject):
    # We need to scope the attribute value, since the for loop changes it.
    def lazy_format(attribute):
        return lambda: format(attribute, tags=True)

    return CssNode(
        "compartment",
        subject,
        Box(
            *(
                CssNode(
                    "attribute",
                    attribute,
                    Text(
                        text=lazy_format(attribute),
                    ),
                )
                for attribute in subject.ownedAttribute
                if not attribute.association
            ),
            draw=draw_top_separator,
        ),
    )


def operations_compartment(subject):
    def lazy_format(operation):
        return lambda: format(
            operation, visibility=True, type=True, multiplicity=True, default=True
        )

    return CssNode(
        "compartment",
        subject,
        Box(
            *(
                CssNode(
                    "operation",
                    operation,
                    Text(
                        text=lazy_format(operation),
                    ),
                )
                for operation in subject.ownedOperation
            ),
            draw=draw_top_separator,
        ),
    )



"""This module contains the actions used in the Toolbox (lower left section of
the main window).

We bind the Toolbox to a diagram. When a diagram page (tab) is switched,
the actions bound to the toolbuttons should change as well.
"""

from collections.abc import Callable, Collection, Sequence
from typing import (
    NamedTuple,
    TypeVar,
)

from gaphor.core.modeling import Base, Diagram, Presentation
from gaphor.diagram.group import group

ItemFactory = Callable[[Diagram, Presentation | None], Presentation]
P = TypeVar("P", bound=Presentation, covariant=True)
ConfigFuncType = Callable[[P], None]


class ToolDef(NamedTuple):
    id: str
    name: str
    icon_name: str
    shortcut: str | None
    item_factory: ItemFactory | None
    handle_index: int = -1


class ToolSection(NamedTuple):
    name: str
    tools: Collection[ToolDef]


ToolboxDefinition = Sequence[ToolSection]


class DiagramType:
    id: str
    name: str
    diagram_type: type[Diagram] | None
    sections: Collection[ToolSection]

    def __init__(
        self,
        diagram_type: type[Diagram],
        name: str,
        sections: tuple[ToolSection, ...],
        allowed_owner_types: tuple[type[None | Base], ...] = (
            type(None),
            Base,
        ),
    ):
        self.id = diagram_type.diagramType.default or ""
        self.diagram_type = diagram_type
        self.name = name
        self.sections = sections
        self.allowed_owner_types = allowed_owner_types

    def allowed(self, element: Base | None) -> bool:
        return isinstance(element, self.allowed_owner_types)

    def create(self, element_factory, element):
        diagram = element_factory.create(self.diagram_type)
        diagram.element = element
        diagram.name = diagram.gettext(self.name)
        diagram.diagramType = self.id

        return diagram


DiagramTypes = Sequence[DiagramType]


class ElementCreateInfo(NamedTuple):
    id: str
    name: str
    element_type: type[Base]
    allowed_owning_elements: tuple[type[Base], ...]


def tooliter(toolbox_actions: Sequence[tuple[str, Sequence[ToolDef]]]):
    """Iterate toolbox items, regardless of section headers."""
    for _name, section in toolbox_actions:
        yield from section


def get_tool_def(modeling_language, tool_name):
    return next(
        t for t in tooliter(modeling_language.toolbox_definition) if t.id == tool_name
    )


def new_item_factory(
    item_class: type[Presentation],
    subject_class: type[Base] | None = None,
    config_func: ConfigFuncType | None = None,
):
    """``config_func`` may be a function accepting the newly created item."""

    def item_factory(diagram, parent=None):
        if subject_class:
            element_factory = diagram.model
            subject = element_factory.create(subject_class)
        else:
            subject = None

        item = diagram.create(item_class, subject=subject)

        if parent and subject and group(parent.subject, item.subject):
            item.change_parent(parent)

        if config_func:
            config_func(item)

        return item

    item_factory.item_class = item_class  # type: ignore[attr-defined]
    item_factory.subject_class = subject_class  # type: ignore[attr-defined]
    return item_factory



from math import pi

from gaphor.C4Model import c4model
from gaphor.diagram.presentation import ElementPresentation, Named, text_name
from gaphor.diagram.shapes import Box, CssNode, Text, stroke
from gaphor.diagram.support import represents


@represents(c4model.Person)
class C4PersonItem(Named, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=48, height=48)

        self.watch("subject[UML:NamedElement].name")
        self.watch("subject[Person].description")

    def update_shapes(self, event=None):
        self.shape = Box(
            text_name(self),
            CssNode(
                "technology",
                self.subject,
                Text(
                    text=lambda: f"[{self.diagram.gettext('Person')}]",
                ),
            ),
            CssNode(
                "description",
                self.subject,
                Text(
                    text=lambda: self.subject.description or "",
                ),
            ),
            draw=draw_person,
        )


def draw_person(box, context, bounding_box):
    cr = context.cairo
    d = context.style.get("border-radius", 24)
    x, y, width, height = bounding_box

    x1 = width + x
    y1 = height + y
    d_head = width * 0.19
    cr.move_to(x, d)
    cr.arc(d, d, d, pi, 1.5 * pi)
    cr.line_to(x + width / 2 - d_head / 2, y)
    cr.arc(x + width / 2, -d_head * 0.866, d_head, 2 / 3 * pi, 1 / 3 * pi)
    cr.line_to(x1 - d, y)
    cr.arc(x1 - d, d, d, 1.5 * pi, y)
    cr.line_to(x1, y1 - d)
    cr.arc(x1 - d, y1 - d, d, 0, 0.5 * pi)
    cr.line_to(d, y1)
    cr.arc(d, y1 - d, d, 0.5 * pi, pi)

    cr.close_path()

    stroke(context, fill=True)



# This file is generated by coder.py. DO NOT EDIT!
# ruff: noqa: F401, E402, F811
# fmt: off

from __future__ import annotations

from gaphor.core.modeling.properties import (
    association,
    attribute as _attribute,
    derived,
    derivedunion,
    enumeration as _enumeration,
    redefine,
    relation_many,
    relation_one,
)


# 1: override Base
from gaphor.core.modeling.base import Base

# 7: override Diagram
from gaphor.core.modeling.diagram import Diagram

# 10: override Presentation
from gaphor.core.modeling.presentation import Presentation

# 16: override StyleSheet
from gaphor.core.modeling.stylesheet import StyleSheet

class PendingChange(Base):
    applied: _attribute[int] = _attribute("applied", int, default=0)
    element_id: _attribute[str] = _attribute("element_id", str)
    op = _enumeration("op", ("add", "remove", "update"), "add")


class ElementChange(PendingChange):
    diagram_id: _attribute[str] = _attribute("diagram_id", str)
    element_name: _attribute[str] = _attribute("element_name", str)
    modeling_language: _attribute[str] = _attribute("modeling_language", str)


class ValueChange(PendingChange):
    property_name: _attribute[str] = _attribute("property_name", str)
    property_type: _attribute[str] = _attribute("property_type", str)
    property_value: _attribute[str] = _attribute("property_value", str)


class RefChange(PendingChange):
    property_name: _attribute[str] = _attribute("property_name", str)
    property_ref: _attribute[str] = _attribute("property_ref", str)



Base.presentation = association("presentation", Presentation, composite=True, opposite="subject")
Diagram.ownedPresentation = association("ownedPresentation", Presentation, composite=True, opposite="diagram")
Presentation.parent = association("parent", Presentation, upper=1, opposite="children")
Presentation.children = association("children", Presentation, composite=True, opposite="parent")
Presentation.diagram = association("diagram", Diagram, upper=1, opposite="ownedPresentation")
Presentation.subject = association("subject", Base, upper=1, opposite="presentation")



"""Profile Import dependency relationship."""

from gaphor import UML
from gaphor.diagram.presentation import LinePresentation
from gaphor.diagram.shapes import draw_arrow_head
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_stereotypes


@represents(
    UML.PackageImport,
    head=UML.PackageImport.importedPackage,
    tail=UML.PackageImport.importingNamespace,
)
class PackageImportItem(LinePresentation):
    """Profile Import dependency relationship."""

    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape_middle=text_stereotypes(
                self, lambda: [self.diagram.gettext("import")]
            ),
        )

        self._handles[0].pos = (30, 20)
        self._handles[1].pos = (0, 0)

        self.watch("subject[Element].appliedStereotype.classifier.name")
        self.draw_head = draw_arrow_head



from __future__ import annotations

from pathlib import Path

from gi.repository import Gtk

from gaphor.abc import ActionProvider, Service
from gaphor.core import action, event_handler, gettext
from gaphor.diagram.event import DiagramClosed, DiagramOpened
from gaphor.diagram.export import (
    escape_filename,
    save_eps,
    save_pdf,
    save_png,
    save_svg,
)
from gaphor.event import ActionEnabled
from gaphor.plugins.diagramexport.exportall import export_all
from gaphor.ui.filedialog import save_file_dialog


class DiagramExport(Service, ActionProvider):
    """Service for exporting diagrams as images (SVG, PNG, PDF)."""

    def __init__(
        self,
        event_manager,
        diagrams=None,
        export_menu=None,
        main_window=None,
        element_factory=None,
    ):
        super().__init__()
        self.event_manager = event_manager
        self.diagrams = diagrams
        self.export_menu = export_menu
        self.main_window = main_window
        if export_menu:
            export_menu.add_actions(self)
        self.filename: Path = Path("export").absolute()
        self.factory = element_factory
        event_manager.subscribe(self.on_diagram_opened_or_closed)

    def shutdown(self):
        self.event_manager.unsubscribe(self.on_diagram_opened_or_closed)
        if self.export_menu:
            self.export_menu.remove_actions(self)

    async def save_dialog(self, diagram, title, ext, mime_type, handler):
        dot_ext = f".{ext}"
        filename = self.filename.with_name(
            escape_filename(diagram.name) or "export"
        ).with_suffix(dot_ext)

        new_filename = await save_file_dialog(
            title,
            filename,
            parent=self.main_window.window,
            filters=[
                (
                    gettext("All {ext} Files").format(ext=ext.upper()),
                    dot_ext,
                    mime_type,
                )
            ],
        )
        if new_filename:
            self.filename = new_filename
            handler(new_filename, diagram)

    @action(
        name="file-export-svg",
        label=gettext("Export as SVG"),
        tooltip=gettext("Export diagram as SVG"),
    )
    async def save_svg_action(self):
        diagram = self.diagrams.get_current_diagram()
        await self.save_dialog(
            diagram, gettext("Export diagram as SVG"), "svg", "image/svg+xml", save_svg
        )

    @action(
        name="file-export-png",
        label=gettext("Export as PNG"),
        tooltip=gettext("Export diagram as PNG"),
    )
    async def save_png_action(self):
        diagram = self.diagrams.get_current_diagram()
        await self.save_dialog(
            diagram, gettext("Export diagram as PNG"), "png", "image/png", save_png
        )

    @action(
        name="file-export-pdf",
        label=gettext("Export as PDF"),
        tooltip=gettext("Export diagram as PDF"),
    )
    async def save_pdf_action(self):
        diagram = self.diagrams.get_current_diagram()
        await self.save_dialog(
            diagram,
            gettext("Export diagram as PDF"),
            "pdf",
            "application/pdf",
            save_pdf,
        )

    @action(
        name="file-export-eps",
        label=gettext("Export as EPS"),
        tooltip=gettext("Export diagram as Encapsulated PostScript"),
    )
    async def save_eps_action(self):
        diagram = self.diagrams.get_current_diagram()
        await self.save_dialog(
            diagram,
            gettext("Export diagram as EPS"),
            "eps",
            "application/postscript",
            save_eps,
        )

    @action(
        name="all-export-svg",
        label=gettext("Export all diagrams as SVG"),
        tooltip=gettext("Export all diagrams as SVG diagrams in a specifieddirectory"),
    )
    def export_all_svg_action(self):
        dialog = Gtk.FileDialog.new()
        dialog.set_title(gettext("Export all diagrams"))

        def response(dialog, result):
            if result.had_error():
                return

            folder = dialog.select_folder_finish(result).get_path()
            export_all(self.factory, folder, save_svg, "svg")

        dialog.select_folder(callback=response)

    @action(
        name="all-export-png",
        label=gettext("Export all diagrams as PNG"),
        tooltip=gettext("Export all diagrams as PNG diagrams in a specifieddirectory"),
    )
    def export_all_png_action(self):
        dialog = Gtk.FileDialog.new()
        dialog.set_title(gettext("Export all diagrams"))

        def response(dialog, result):
            if result.had_error():
                return

            folder = dialog.select_folder_finish(result).get_path()
            export_all(self.factory, folder, save_png, "png")

        dialog.select_folder(callback=response)

    @action(
        name="all-export-pdf",
        label=gettext("Export all diagrams as PDF"),
        tooltip=gettext("Export all diagrams as PDF diagrams in a specifieddirectory"),
    )
    def export_all_pdf_action(self):
        dialog = Gtk.FileDialog.new()
        dialog.set_title(gettext("Export all diagrams"))

        def response(dialog, result):
            if result.had_error():
                return

            folder = dialog.select_folder_finish(result).get_path()
            export_all(self.factory, folder, save_pdf, "pdf")

        dialog.select_folder(callback=response)

    @action(
        name="all-export-eps",
        label=gettext("Export all diagrams as EPS"),
        tooltip=gettext("Export all diagrams as EPS diagrams in a specifieddirectory"),
    )
    def export_all_eps_action(self):
        dialog = Gtk.FileDialog.new()
        dialog.set_title(gettext("Export all diagrams"))

        def response(dialog, result):
            if result.had_error():
                return

            folder = dialog.select_folder_finish(result).get_path()
            export_all(self.factory, folder, save_eps, "eps")

        dialog.select_folder(callback=response)

    @event_handler(DiagramOpened, DiagramClosed)
    def on_diagram_opened_or_closed(self, event: DiagramOpened | DiagramClosed):
        enabled = (
            isinstance(event, DiagramOpened) or self.diagrams.get_current_diagram()
        )

        for action_name in (
            "win.file-export-svg",
            "win.file-export-png",
            "win.file-export-pdf",
            "win.file-export-eps",
        ):
            self.event_manager.handle(ActionEnabled(action_name, enabled))



"""Node item may represent a node or a device UML metamodel classes.

Grouping
========
Node item can group following items

- other nodes, which are represented with Node.nestedNode on UML metamodel
  level
- deployed artifacts using deployment
- components, which are parts of a node acting as structured classifier
  (nodes may have internal structures)

Node item grouping logic is implemented in `gaphor.adapters.grouping`
module.
"""

from gaphor import UML
from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import Classified, ElementPresentation
from gaphor.diagram.shapes import Box, stroke
from gaphor.diagram.support import represents
from gaphor.UML.classes.stereotype import stereotype_compartments, stereotype_watches
from gaphor.UML.compartments import name_compartment


@represents(UML.Node)
@represents(UML.Device)
class NodeItem(Classified, ElementPresentation):
    """Representation of node or device from UML Deployment package."""

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)

        self.watch("children", self.update_shapes)
        self.watch("show_stereotypes", self.update_shapes)
        self.watch("subject[NamedElement].name")
        self.watch("subject[Node].ownedConnector", self.update_shapes)
        stereotype_watches(self)

    show_stereotypes: attribute[int] = attribute("show_stereotypes", int)

    def update_shapes(self, event=None):
        self.shape = Box(
            name_compartment(
                self,
                lambda: (
                    [self.diagram.gettext("device")]
                    if isinstance(self.subject, UML.Device)
                    else []
                ),
            ),
            *(self.show_stereotypes and stereotype_compartments(self.subject) or []),
            draw=draw_node,
        )


def draw_node(box, context, bounding_box):
    cr = context.cairo

    d = 10
    w = bounding_box.width
    h = bounding_box.height

    cr.rectangle(0, 0, w, h)

    cr.move_to(0, 0)
    cr.line_to(d, -d)
    cr.line_to(w + d, -d)
    cr.line_to(w + d, h - d)
    cr.line_to(w, h)
    cr.move_to(w, 0)
    cr.line_to(w + d, -d)

    stroke(context, fill=True)



import gaphor.UML.interactions.drop  # noqa: F401
from gaphor.UML.interactions import (  # noqa: F401
    copypaste,
    interactionsconnect,
    interactionsgroup,
)
from gaphor.UML.interactions.executionspecification import ExecutionSpecificationItem
from gaphor.UML.interactions.interaction import InteractionItem
from gaphor.UML.interactions.lifeline import LifelineItem
from gaphor.UML.interactions.message import MessageItem

__all__ = [
    "ExecutionSpecificationItem",
    "InteractionItem",
    "LifelineItem",
    "MessageItem",
]



import functools
import importlib
import logging

from gaphas.guide import GuidePainter
from gaphas.painter import FreeHandPainter, HandlePainter, PainterChain
from gaphas.segment import LineSegmentPainter
from gaphas.tool.itemtool import find_item_and_handle_at_point
from gaphas.tool.rubberband import RubberbandPainter, RubberbandState
from gaphas.view import GtkView
from gi.repository import Adw, Gdk, GdkPixbuf, Gio, GLib, Gtk

from gaphor.core import event_handler, gettext
from gaphor.core.modeling.diagram import StyledDiagram
from gaphor.core.modeling.event import (
    AttributeUpdated,
    StyleSheetUpdated,
)
from gaphor.diagram.diagramtoolbox import get_tool_def, tooliter
from gaphor.diagram.event import DiagramSelectionChanged
from gaphor.diagram.painter import DiagramTypePainter, ItemPainter
from gaphor.diagram.tools import (
    apply_default_tool_set,
    apply_magnet_tool_set,
    apply_placement_tool_set,
)
from gaphor.diagram.tools.magnet import MagnetPainter
from gaphor.i18n import translated_ui_string
from gaphor.transaction import Transaction
from gaphor.ui.actiongroup import apply_action_group
from gaphor.ui.clipboard import Clipboard
from gaphor.ui.event import ToolSelected

log = logging.getLogger(__name__)


@functools.lru_cache(maxsize=1)
def placement_icon_base():
    f = importlib.resources.files("gaphor") / "ui" / "placement-icon-base.png"
    return GdkPixbuf.Pixbuf.new_from_file_at_scale(str(f), 64, 64, True)


if hasattr(GtkView, "set_css_name"):
    GtkView.set_css_name("diagramview")


def new_builder():
    builder = Gtk.Builder()
    builder.add_from_string(translated_ui_string("gaphor.ui", "diagrampage.ui"))
    return builder


@functools.cache
def get_placement_icon(display, icon_name):
    if display is None:
        display = Gdk.Display.get_default()
    pixbuf = placement_icon_base().copy()
    theme_icon = Gtk.IconTheme.get_for_display(display).lookup_icon(
        icon_name,
        None,
        24,
        1,
        Gtk.TextDirection.NONE,
        Gtk.IconLookupFlags.FORCE_SYMBOLIC,
    )
    icon = GdkPixbuf.Pixbuf.new_from_file_at_scale(
        theme_icon.get_file().get_path(), 32, 32, True
    )
    icon.copy_area(
        0,
        0,
        icon.get_width(),
        icon.get_height(),
        pixbuf,
        9,
        15,
    )
    return Gdk.Texture.new_for_pixbuf(pixbuf)


def get_placement_cursor(display, icon_name):
    return Gdk.Cursor.new_from_texture(get_placement_icon(display, icon_name), 1, 1)


class DiagramPage:
    def __init__(self, diagram, event_manager, element_factory, modeling_language):
        self.event_manager = event_manager
        self.diagram = diagram
        self.modeling_language = modeling_language
        self.clipboard = Clipboard(event_manager, element_factory)
        self.style_manager = Adw.StyleManager.get_default()

        self.view: GtkView | None = None
        self.alignment_button: Gtk.Button | None = None
        self.diagram_css: Gtk.CssProvider | None = None

        self.rubberband_state = RubberbandState()
        self.context_menu = Gtk.PopoverMenu.new_from_model(popup_model(diagram))

        self.event_manager.subscribe(self._on_attribute_updated)
        self.event_manager.subscribe(self._on_style_sheet_updated)
        self.event_manager.subscribe(self._on_tool_selected)

    @property
    def title(self):
        return self.diagram and self.diagram.name or gettext("<None>")

    def get_diagram(self):
        return self.diagram

    def get_view(self):
        return self.view

    def construct(self):
        """Create the widget.

        Returns: the newly created widget.
        """
        assert self.diagram

        builder = new_builder()
        view: GtkView = builder.get_object("view")
        view.add_css_class(self._css_class())
        view.connect("delete", delete_selected_items, self.event_manager)
        view.connect("cut-clipboard", self.clipboard.cut)
        view.connect("copy-clipboard", self.clipboard.copy)
        view.connect("paste-clipboard", self.clipboard.paste_link)
        view.connect("paste-full-clipboard", self.clipboard.paste_full)
        view.selection.add_handler(self._selection_changed)
        self.clipboard.clipboard.connect(
            "notify::content", self._clipboard_content_changed
        )

        self.diagram_css = Gtk.CssProvider.new()
        Gtk.StyleContext.add_provider_for_display(
            Gdk.Display.get_default(),
            self.diagram_css,
            Gtk.STYLE_PROVIDER_PRIORITY_USER,
        )

        self.style_manager.connect("notify::dark", self._on_notify_dark)

        self.view = view
        self.context_menu.set_parent(view)

        self.select_tool("toolbox-pointer")

        self.update_drawing_style()

        # Set model only after the painters are set
        view.model = self.diagram

        diagrampage = builder.get_object("diagrampage")
        apply_action_group(self, "diagram", diagrampage)
        self._clipboard_content_changed(self.clipboard.clipboard)

        self.alignment_button = builder.get_object("alignment-button")

        return diagrampage

    def apply_tool_set(self, tool_name):
        """Return a tool associated with an id (action name).

        Tool sets can't be changed if a transaction is active,
        because it could ruin the user (inter)action.
        """

        if Transaction.in_transaction():
            log.warning("Cannot change tool set while in a transaction.")
            return False

        if tool_name == "toolbox-pointer":
            apply_default_tool_set(
                self.view,
                self.modeling_language,
                self.event_manager,
                self.rubberband_state,
            )
            if self.view:
                self.view.add_controller(
                    context_menu_controller(self.context_menu, self.diagram)
                )

        elif tool_name == "toolbox-magnet":
            apply_magnet_tool_set(
                self.view,
                self.modeling_language,
                self.event_manager,
            )
        else:
            tool_def = get_tool_def(self.modeling_language, tool_name)
            item_factory = tool_def.item_factory
            handle_index = tool_def.handle_index
            apply_placement_tool_set(
                self.view,
                item_factory=item_factory,
                modeling_language=self.modeling_language,
                event_manager=self.event_manager,
                handle_index=handle_index,
            )
        return True

    def get_tool_icon_name(self, tool_name):
        if tool_name == "toolbox-pointer":
            return None
        return next(
            t
            for t in tooliter(self.modeling_language.toolbox_definition)
            if t.id == tool_name
        ).icon_name

    def _css_class(self):
        return f"diagram-{id(self)}"

    def _selection_changed(self, _item):
        view = self.view
        assert view
        selection = view.selection
        self.event_manager.handle(
            DiagramSelectionChanged(
                view, selection.focused_item, selection.selected_items
            )
        )

    @event_handler(ToolSelected)
    def _on_tool_selected(self, event: ToolSelected):
        self.select_tool(event.tool_name)

    @event_handler(AttributeUpdated)
    def _on_attribute_updated(self, event: AttributeUpdated):
        if event.property.name == "name" and self.view:
            self.view.update_back_buffer()

    @event_handler(StyleSheetUpdated)
    def _on_style_sheet_updated(self, _event):
        self.update_drawing_style()
        self.diagram.update(self.diagram.ownedPresentation)

    def _clipboard_content_changed(self, clipboard, _pspec=None):
        if not self.view:
            return

        enabled = self.clipboard.can_paste()
        self.view.action_set_enabled("clipboard.paste", enabled)
        self.view.action_set_enabled("clipboard.paste-full", enabled)

    def _on_notify_dark(self, style_manager, gparam):
        self.update_drawing_style()

    def close(self):
        """Tab is destroyed.

        Do the same thing that would be done if Close was pressed.
        """
        assert self.view

        Gtk.StyleContext.remove_provider_for_display(
            Gdk.Display.get_default(),
            self.diagram_css,
        )

        self.event_manager.unsubscribe(self._on_attribute_updated)
        self.event_manager.unsubscribe(self._on_style_sheet_updated)
        self.event_manager.unsubscribe(self._on_tool_selected)
        self.view = None

    def select_tool(self, tool_name: str):
        if not self.view:
            return

        if self.apply_tool_set(tool_name):
            if icon_name := self.get_tool_icon_name(tool_name):
                self.view.set_cursor(get_placement_cursor(None, icon_name))
            else:
                self.view.set_cursor(None)

    def update_drawing_style(self):
        """Set the drawing style for the diagram based on the active style
        sheet."""
        assert self.view
        assert self.diagram_css

        dark_mode = self.style_manager.get_dark()
        style = self.diagram.style(StyledDiagram(self.diagram, dark_mode=dark_mode))

        bg = style.get("background-color", (0.0, 0.0, 0.0, 0.0))
        self.diagram_css.load_from_string(
            f".{self._css_class()} {{ background-color: rgba({int(255 * bg[0])}, {int(255 * bg[1])}, {int(255 * bg[2])}, {bg[3]}); }}",
        )

        view = self.view
        item_painter = ItemPainter(view.selection, dark_mode)

        if sloppiness := style.get("line-style", 0.0):
            item_painter = FreeHandPainter(item_painter, sloppiness=sloppiness)

        view.bounding_box_painter = item_painter
        view.painter = (
            PainterChain()
            .append(item_painter)
            .append(HandlePainter(view))
            .append(LineSegmentPainter(view.selection))
            .append(GuidePainter(view))
            .append(MagnetPainter(view))
            .append(RubberbandPainter(self.rubberband_state))
            .append(DiagramTypePainter(self.diagram))
        )

        view.request_update(self.diagram.get_all_items())


def delete_selected_items(view: GtkView, event_manager):
    with Transaction(event_manager):
        items = view.selection.selected_items
        for i in list(items):
            i.unlink()


def context_menu_controller(context_menu, diagram):
    def on_show_popup(ctrl, n_press, x, y):
        if (
            Transaction.in_transaction()
            or ctrl.get_last_event() is None
            or not ctrl.get_last_event().triggers_context_menu()
        ):
            return

        view = ctrl.get_widget()
        item, _handle = find_item_and_handle_at_point(view, (x, y))

        context_menu.set_menu_model(
            popup_model(item.subject if item and item.subject else diagram)
        )

        gdk_rect = Gdk.Rectangle()
        gdk_rect.x = x
        gdk_rect.y = y
        gdk_rect.width = gdk_rect.height = 1

        context_menu.set_has_arrow(False)
        context_menu.set_pointing_to(gdk_rect)
        context_menu.popup()

    ctrl = Gtk.GestureClick.new()
    ctrl.set_button(0)
    ctrl.connect("pressed", on_show_popup)
    return ctrl


def popup_model(element):
    model = Gio.Menu.new()
    part = Gio.Menu.new()

    menu_item = Gio.MenuItem.new(
        gettext("Show in Model Browser"),
        "win.show-in-model-browser",
    )
    menu_item.set_attribute_value("target", GLib.Variant.new_string(element.id))

    part.append_item(menu_item)
    model.append_section(None, part)

    return model



from gaphor.diagram.presentation import (
    Classified,
    ElementPresentation,
)
from gaphor.diagram.shapes import (
    DEFAULT_PADDING,
    Box,
    CssNode,
    Orientation,
    Text,
    cairo_state,
    draw_border,
    stroke,
)
from gaphor.SysML.sysml import Block, ConstraintBlock, Requirement
from gaphor.UML import Interaction, Package, Profile, StateMachine
from gaphor.UML.actions.activity import update_parameter_nodes
from gaphor.UML.recipes import stereotypes_str
from gaphor.UML.uml import Activity, Diagram, NamedElement


class DiagramFrameItem(Classified, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape=Box(
                Box(
                    CssNode(
                        "pentagon",
                        self,
                        Box(
                            CssNode(
                                "diagramtype",
                                None,
                                Text(text=lambda: self.diagram.diagramType),
                            ),
                            CssNode(
                                "stereotypes",
                                None,
                                Text(text=lambda: stereotypes_str(diagram)),
                            ),
                            CssNode("name", None, Text(text=self.label)),
                            orientation=Orientation.HORIZONTAL,
                            draw=draw_pentagon,
                        ),
                    )
                ),
                draw=draw_border,
            ),
        )

        self.watch("subject[UML:NamedElement].name").watch(
            "diagram[UML:Diagram].name"
        ).watch("diagram[UML:Diagram].element[UML:NamedElement].name").watch(
            "subject[UML:Activity].node", self.update_activity_parameters
        )

    def label(self):
        assert isinstance(self.diagram, Diagram)
        return diagram_label(self.diagram)

    def postload(self):
        super().postload()
        if isinstance(self.subject, Activity):
            self.update_activity_parameters()

    def update_activity_parameters(self, event=None):
        if not isinstance(self.subject, Activity):
            return

        update_parameter_nodes(self)


def draw_pentagon(box, context, _bounding_box):
    cr = context.cairo
    _padding_top, _padding_right, padding_bottom, padding_left = context.style.get(
        "padding", DEFAULT_PADDING
    )
    w = sum(x for x, _y in box.sizes) + padding_left
    h = max(y for _x, y in box.sizes) + padding_bottom

    with cairo_state(cr):
        h2 = h / 2.0
        cr.move_to(0, h)
        cr.line_to(w - 4, h)
        cr.line_to(w, h2)
        cr.line_to(w, 0)
        stroke(context, fill=False)


def diagram_label(diagram):
    if (el := diagram.element) and isinstance(el, NamedElement):
        return f"[{_model_element_type(el)}] {el.name} [{diagram.name}]"

    # TODO: SysML specification does not allow parentless elements,
    #       but since it is not constrained (yet), it may happen.
    return diagram.name


def _model_element_type(el) -> str:
    if isinstance(el, Activity):
        return "activity"
    if isinstance(el, ConstraintBlock):
        return "constraint block"
    if isinstance(el, Block):
        return "block"
    if isinstance(el, Interaction):
        return "interaction"
    if isinstance(el, Profile):
        return "profile"
    if isinstance(el, Package):
        return "package"
    if isinstance(el, Requirement):
        return "requirement"
    return "state machine" if isinstance(el, StateMachine) else ""



from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager
from typing import Protocol
from xml.sax.saxutils import escape, quoteattr

# See whether the xmlcharrefreplace error handler is
# supported
try:
    from codecs import xmlcharrefreplace_errors

    _error_handling = "xmlcharrefreplace"
    del xmlcharrefreplace_errors
except ImportError:
    _error_handling = "strict"


class WriterProtocol(Protocol):
    def write(self, text: str):
        """Write text"""


class XMLWriter:
    def __init__(self, out: WriterProtocol, encoding="utf-8"):
        super().__init__()
        self._out = out
        self._encoding = encoding
        self._current_context: dict[str, str] = {}  # contains uri -> prefix mapping
        self._undeclared_ns_maps: list[tuple[str, str]] = []

        self._in_cdata = False
        self._in_start_tag = False
        self._next_newline = False

    def _write(self, text: str, start_tag=False, end_tag=False) -> None:
        """Write data.

        Tags should not be escaped. They should be marked by setting
        either ``start_tag`` or ``end_tag`` to ``True``. Only the tag
        should be marked this way. Other stuff, such as namespaces and
        attributes can be written directly to the file.
        """
        if not isinstance(text, str):
            text = text.decode(self._encoding, _error_handling)

        if self._next_newline:
            self._out.write("\n")
            self._next_newline = False

        if start_tag and not self._in_start_tag:
            self._in_start_tag = True
            self._out.write("<")
        elif start_tag:
            self._out.write(">")
            self._out.write("\n")
            self._out.write("<")
        elif end_tag and self._in_start_tag:
            self._out.write("/>")
            self._in_start_tag = False
            self._next_newline = True
            return
        elif self._in_start_tag:
            self._out.write(">")
            self._in_start_tag = False
        elif end_tag:
            self._out.write("</")
            self._out.write(text)
            self._out.write(">")
            self._in_start_tag = False
            self._next_newline = True
            return

        self._out.write(text)

    def _qname(self, name: tuple[str, str]) -> str:
        """Builds a qualified name from a (ns_url, localname) pair."""
        if name[0]:
            if prefix := self._current_context[name[0]]:
                # If it is not the default namespace, prepend the prefix
                return f"{prefix}:{name[1]}"
        # Return the unqualified name
        return name[1]

    @contextmanager
    def document(self) -> Iterator[XMLWriter]:
        self._write(f'<?xml version="1.0" encoding="{self._encoding}"?>\n')
        yield self

    @contextmanager
    def element_ns(
        self, name: tuple[str, str], attrs: dict[tuple[str, str], str]
    ) -> Iterator[XMLWriter]:
        with self._prefix_mappings() as xmlns:
            self._write(self._qname(name), start_tag=True)
            self._out.write(xmlns)

            for attr, value in list(attrs.items()):
                self._out.write(f" {self._qname(attr)}={quoteattr(value)}")

            yield self

            self._write(f"{self._qname(name)}", end_tag=True)

    @contextmanager
    def element(self, name: str, attrs: dict[str, str]) -> Iterator[XMLWriter]:
        """Create new element in default namespace."""
        ns = next(
            (uri for uri, prefix in self._current_context.items() if prefix == ""), ""
        )
        with self.element_ns(
            (ns, name), {(ns, name): value for name, value in attrs.items()}
        ):
            yield self

    @contextmanager
    def cdata(self) -> Iterator[XMLWriter]:
        self._write("<![CDATA[")
        self._in_cdata = True

        yield self

        self._write("]]>")
        self._in_cdata = False

    def prefix_mapping(self, prefix: str, uri: str) -> None:
        self._undeclared_ns_maps.append((prefix, uri))

    @contextmanager
    def _prefix_mappings(self) -> Iterator[str]:
        if not self._undeclared_ns_maps:
            yield ""
        else:
            current_context = self._current_context.copy()

            xmlns = ""
            for prefix, uri in self._undeclared_ns_maps:
                self._current_context[uri] = prefix
                if prefix:
                    xmlns += f' xmlns:{prefix}="{uri}"'
                else:
                    xmlns += f' xmlns="{uri}"'

            self._undeclared_ns_maps = []

            yield xmlns

            self._current_context = current_context

    def characters(self, content: str) -> None:
        if self._in_cdata:
            self._write(content.replace("]]>", "] ]>"))
        else:
            self._write(escape(content))

    def processing_instruction(self, target: str, data: str) -> None:
        self._write(f"<?{target} {data}?>")

    def comment(self, comment: str) -> None:
        self._write("<!-- ")
        self._write(comment.replace("-->", "- ->"))
        self._write(" -->")



"""GUI Undo/redo actions

Undo and redo are special, since they share their
action name with text widget undo actions.

This service makes sure the text edit undo/redo actions
have preference over window level undo.

Undo/redo have to use the same action names, so
the functions are accessible from the (macOS)
menu bar.
"""

from gaphor.abc import ActionProvider, Service
from gaphor.action import action
from gaphor.services.undomanager import UndoManager
from gaphor.ui.mainwindow import MainWindow


class UndoActions(Service, ActionProvider):
    def __init__(self, undo_manager: UndoManager, main_window: MainWindow):
        self.undo_manager = undo_manager
        self.main_window = main_window
        self.in_action = False

    def shutdown(self):
        pass

    @action(name="text.undo", shortcut="<Primary>z")
    def undo_transaction(self):
        self.activate_on_focused_widget("text.undo", self.undo_manager.undo_transaction)

    @action(name="text.redo", shortcut="<Primary><Shift>z")
    def redo_transaction(self):
        self.activate_on_focused_widget("text.redo", self.undo_manager.redo_transaction)

    def activate_on_focused_widget(self, action_name, func):
        if self.in_action:
            return func()

        self.in_action = True
        try:
            if not (
                (win := self.main_window.window)
                and (widget := win.get_focus())
                and widget.activate_action(action_name)
            ):
                func()
        finally:
            self.in_action = False



"""Interaction diagram item."""

from gaphor import UML
from gaphor.diagram.presentation import ElementPresentation, Named, text_name
from gaphor.diagram.shapes import Box, CssNode, stroke
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_stereotypes


@represents(UML.Interaction)
class InteractionItem(Named, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=150, height=100)

        self.shape = Box(
            CssNode(
                "pentagon",
                None,
                Box(
                    text_stereotypes(self),
                    text_name(self),
                ),
            ),
            draw=draw_interaction,
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject[Element].appliedStereotype.classifier.name")


def draw_interaction(box, context, bounding_box):
    cr = context.cairo
    cr.rectangle(0, 0, bounding_box.width, bounding_box.height)
    stroke(context, fill=True)
    # draw pentagon
    if not box.sizes:
        box.size(context, bounding_box)
    w, h = box.sizes[0]
    h2 = h / 2.0
    cr.move_to(0, h)
    cr.line_to(w - 4, h)
    cr.line_to(w, h2)
    cr.line_to(w, 0)
    stroke(context, fill=False)



from gaphor.core.format import format
from gaphor.i18n import gettext
from gaphor.SysML import sysml


@format.register(sysml.Allocate)
@format.register(sysml.Refine)
@format.register(sysml.Trace)
def format_directed_relationship_property_path(el):
    return gettext("sourceContext: {name}").format(
        name=el.sourceContext and el.sourceContext.name or ""
    )
